import itertools
import math
import re
from collections import defaultdict
from queue import PriorityQueue
from time import time

import networkx
import nltk
import numpy as np
import pandas as pd
import tqdm
from tqdm.autonotebook import tqdm

from .datamodel import Data, PYJEDAIFeature
from .evaluation import Evaluation

class AbstractJoin(PYJEDAIFeature):
    """Abstract class of Joins module
    """

    def __init__(
            self,
            metric: str,
            tokenization: str,
            qgrams: int = 2,
            similarity_threshold: float = None
    ) -> None:
        """AbstractJoin Constructor

        Args:
            metric (str): String similarity metric
            tokenization (str): Tokenizer
            qgrams (int, optional): For Jaccard metric. Defaults to 2.
            similarity_threshold (float, optional): Threshold for preserving pair. Defaults to None.
        """
        self.metric = metric
        self.qgrams = qgrams
        self.tokenization = tokenization
        self._source_frequency: np.array
        self.similarity_threshold: float = similarity_threshold
        self.reverse_order: bool
        self.attributes_1: list
        self.attributes_2: list
        self._flags: np.array
        self.pairs: networkx.Graph
        self.vectorizer = None

    def fit(self,
            data: Data,
            reverse_order: bool = False,
            attributes_1: list = None,
            attributes_2: list = None,
            tqdm_disable: bool = False
    ) -> networkx.Graph:
        """Joins main method

            Args:
                data (Data): dataset module
                reverse_order (bool, optional): _description_. Defaults to False.
                attributes_1 (list, optional): _description_. Defaults to None.
                attributes_2 (list, optional): _description_. Defaults to None.
                tqdm_disable (bool, optional): _description_. Defaults to False.

            Returns:
                networkx.Graph: graph containg nodes as entities and edges as similarity score
        """
        if reverse_order and data.is_dirty_er:
            raise ValueError("Can't have reverse order in Dirty Entity Resolution")

        start_time = time()
        self.tqdm_disable, self.reverse_order, self.attributes_1, self.attributes_2, self.data = \
            tqdm_disable, reverse_order, attributes_1, attributes_2, data

        self._entities_d1 = data.dataset_1[attributes_1 if attributes_1 else data.attributes_1] \
                            .apply(" ".join, axis=1) \
                            .apply(self._tokenize_entity) \
                            .values.tolist()

        if not data.is_dirty_er:
            self._entities_d2 = data.dataset_2[attributes_2 if attributes_2 else data.attributes_2] \
                    .apply(" ".join, axis=1) \
                    .apply(self._tokenize_entity) \
                    .values.tolist()

        num_of_entities = self.data.num_of_entities_2 if reverse_order else self.data.num_of_entities_1

        self._progress_bar = tqdm(
            total=self.data.num_of_entities if not self.data.is_dirty_er else num_of_entities*2,
            desc=self._method_name+" ("+self.metric+")", disable=self.tqdm_disable
        )

        self._flags, \
        self._counters, \
        self._sims, \
        self._source_frequency, \
        self.pairs = np.empty([num_of_entities]), \
                    np.zeros([num_of_entities]), \
                    np.empty([self.data.num_of_entities_1*self.data.num_of_entities_2]), \
                    np.empty([num_of_entities]), \
                    networkx.Graph()
        self._flags[:] = -1
        entity_index = self._create_entity_index(
                self._entities_d2 if reverse_order else self._entities_d1
            )

        if self.data.is_dirty_er:
            eid = 0
            for entity in self._entities_d1:
                candidates = set()
                for token in entity:
                    if token in entity_index:
                        current_candidates = entity_index[token]
                        for candidate_id in current_candidates:
                            if self._flags[candidate_id] != eid:
                                self._counters[candidate_id] = 0
                                self._flags[candidate_id] = eid
                            self._counters[candidate_id] += 1
                            candidates.add(candidate_id)
                self._process_candidates(candidates, eid, len(entity))
                self._progress_bar.update(1)
                eid += 1
        else:
            if reverse_order:
                entities = self._entities_d1
                num_of_entities = self.data.num_of_entities_1
            else:
                entities = self._entities_d2
                num_of_entities = self.data.num_of_entities_2

            for i in range(0, num_of_entities):
                candidates = set()
                record = entities[i]
                entity_id = i if reverse_order else i+self.data.dataset_limit
                for token in record:
                    if token in entity_index:
                        current_candidates = entity_index[token]
                        for candidate_id in current_candidates:
                            if self._flags[candidate_id] != entity_id:
                                self._counters[candidate_id] = 0
                                self._flags[candidate_id] = entity_id
                            self._counters[candidate_id] += 1
                            candidates.add(candidate_id)
                if 0 < len(candidates):
                    self._process_candidates(candidates, entity_id, len(record))
                self._progress_bar.update(1)
        self._progress_bar.close()
        self.execution_time = time() - start_time
        return self.pairs

    def _tokenize_entity(self, entity: str) -> set:
        if self.vectorizer is not None:
            return entity.lower()
        elif self.tokenization == 'qgrams':
            return set([' '.join(grams) for grams in nltk.ngrams(entity.lower(), n=self.qgrams)])
        elif self.tokenization == 'standard':
            return set(filter(None, re.split('[\\W_]', entity.lower())))
        elif self.tokenization == 'standard_multiset':
            tok_ids_index = {}
            multiset = set()
            for tok in set(filter(None, re.split('[\\W_]', entity.lower()))):
                tok_id =  tok_ids_index[tok] if tok in tok_ids_index else 0
                multiset.add(tok+str(tok_id))
                tok_ids_index[tok] = tok_id+1
            return multiset
        elif self.tokenization == 'qgrams_multiset':
            grams_ids_index = {}
            qgrams = set()
            for gram in set([' '.join(grams) for grams in nltk.ngrams(entity.lower(), n=self.qgrams)]):
                gram_id =  grams_ids_index[gram] if gram in grams_ids_index else 0
                qgrams.add(gram+str(gram_id))
                grams_ids_index[gram] = gram_id+1
            return qgrams
        else:
            raise AttributeError("Tokenization not found")

    def _calc_similarity(
            self,
            common_tokens: int,
            source_frequency: int,
            tokens_size: int
        ) -> float:
        """Similarity score

        Args:
            common_tokens (int): number of common tokens
            source_frequency (int): frequency
            tokens_size (int): size of tokens

        Returns:
            float: similarity
        """
        if self.metric == 'cosine':
            return common_tokens / math.sqrt(source_frequency*tokens_size)
        elif self.metric == 'dice':
            return 2 * common_tokens / (source_frequency+tokens_size)
        elif self.metric == 'jaccard':
            return common_tokens / (source_frequency+tokens_size-common_tokens)
        
    def _calc_vector_similarity(self, id1 : int, id2 : int) -> float:
        """Vector based similarity score

        Args:
            id1 (int): D1 entity ID
            id2 (int): D2 entity ID

        Returns:
            float: vector based similarity
        """
        return self.vectorizer.predict(id1=id1, id2=id2)

    def _create_entity_index(self, entities: list) -> dict:
        entity_index = defaultdict(set)
        entity_id = itertools.count()
        for entity in entities:
            eid = next(entity_id)
            for token in entity:
                entity_index[token].add(eid)
            self._source_frequency[eid] = len(entity)
            self._progress_bar.update(1)

        return entity_index

#     def _similarity(self, entity_id1: int, entity_id2: int, attributes: any=None) -> float:
#         similarity: float = 0.0
#         if isinstance(attributes, dict):
#             for attribute, weight in self.attributes.items():
#                 similarity += weight*self._metric(
#                     self.data.entities.iloc[entity_id1][attribute],
#                     self.data.entities.iloc[entity_id2][attribute]
#                 )
#         if isinstance(attributes, list):
#             for attribute in self.attributes:
#                 similarity += self._metric(
#                     self.data.entities.iloc[entity_id1][attribute],
#                     self.data.entities.iloc[entity_id2][attribute]
#                 )
#                 similarity /= len(self.attributes)
#         else:
#             # print(self.data.entities.iloc[entity_id1].str.cat(sep=' '),
#                 # self.data.entities.iloc[entity_id2].str.cat(sep=' '))
#             # concatenated row string
#             similarity = self._metric(
#                 self.data.entities.iloc[entity_id1].str.cat(sep=' '),
#                 self.data.entities.iloc[entity_id2].str.cat(sep=' ')
#             )
#         return similarity

    def _insert_to_graph(self, entity_id1, entity_id2, similarity):
        if self.similarity_threshold <= similarity:
            self.pairs.add_edge(entity_id1, entity_id2, weight=similarity)

    def evaluate(self, prediction=None, export_to_df: bool = False,
                 export_to_dict: bool = False, with_classification_report: bool = False,
                 verbose: bool = True) -> any:
        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " +
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        total_matching_pairs = prediction.number_of_edges()
        for _, (id1, id2) in self.data.ground_truth.iterrows():
            id1 = self.data._ids_mapping_1[id1]
            id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er \
                                                else self.data._ids_mapping_2[id2]
            if (id1 in prediction and id2 in prediction[id1]) or   \
                 (id2 in prediction and id1 in prediction[id2]):
                true_positives += 1

        eval_obj.calculate_scores(true_positives=true_positives, 
                                  total_matching_pairs=total_matching_pairs)
        return eval_obj.report(self.method_configuration(),
                        export_to_df,
                        export_to_dict,
                        with_classification_report,
                        verbose)
        
    def stats(self) -> None:
        pass

    def _configuration(self) -> dict:
        return {
            "similarity_threshold" : self.similarity_threshold,
            "metric" : self.metric,
            "tokenization" : self.tokenization,
            "qgrams": self.qgrams
        }    

    def export_to_df(self, prediction) -> pd.DataFrame:
        """creates a dataframe with the predicted pairs

        Args:
            prediction (any): Predicted candidate pairs

        Returns:
            pd.DataFrame: Dataframe with the predicted pairs
        """
        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. \
                Data object mush have initialized with the ground-truth file")
        pairs_df = pd.DataFrame(columns=['id1', 'id2'])
        for edge in prediction.edges:
            id1 = self.data._gt_to_ids_reversed_1[edge[0]]
            id2 = self.data._gt_to_ids_reversed_1[edge[1]] if self.data.is_dirty_er \
                        else self.data._gt_to_ids_reversed_2[edge[1]]
            pairs_df = pd.concat([pairs_df, pd.DataFrame([{'id1':id1, 'id2':id2}], index=[0])], ignore_index=True)

        return pairs_df

    
    
    
class EJoin(AbstractJoin):
    """
     E Join algorithm
    """
    _method_name = "EJoin"
    _method_info = " EJoin algorithm"
    _method_short_name = "EJ"

    def __init__(
        self,
        similarity_threshold: float = 0.82,
        metric: str = 'cosine',
        tokenization: str = 'qgrams',
        qgrams: int = 2
    ) -> None:
        super().__init__(metric, tokenization, qgrams, similarity_threshold)

    def _process_candidates(self, candidates: set, entity_id: int, tokens_size: int) -> None:
        for candidate_id in candidates:
            self._insert_to_graph(
                candidate_id+self.data.dataset_limit if self.reverse_order \
                                                        and not self.data.is_dirty_er \
                                                    else candidate_id,
                entity_id,
                self._calc_similarity(
                    self._counters[candidate_id],
                    self._source_frequency[candidate_id],
                    tokens_size
                )
            )

class TopKJoin(AbstractJoin):
    """Top-K Join algorithm
    """

    _method_name = "Top-K Join"
    _method_info = "Top-K Join algorithm"
    _method_short_name = "TopKJ"

    def __init__(self,
                 K: int,
                 metric: str,
                 tokenization: str,
                 qgrams: int = 2
        ) -> None:
        super().__init__(metric, tokenization, qgrams)
        self.K = K

    def _process_candidates(self, candidates: set, entity_id: int, tokens_size: int) -> None:
        minimum_weight=0
        pq = PriorityQueue()
        pq.put(minimum_weight)
        for index, candidate_id in enumerate(candidates):
            if(self.vectorizer is None):
                sim = self._calc_similarity(self._counters[candidate_id], self._source_frequency[candidate_id], tokens_size)
            else:
                sim = self._calc_vector_similarity(((candidate_id + self.data.dataset_limit) if self.reverse_order else candidate_id), entity_id)
            if minimum_weight < sim:
                pq.put(sim)
                if self.K < pq.qsize():
                    minimum_weight = pq.get()

        minimum_weight = pq.get()
        for index, candidate_id in enumerate(candidates):
            self.similarity_threshold = minimum_weight
            if(self.vectorizer is None):
                sim = self._calc_similarity(self._counters[candidate_id], self._source_frequency[candidate_id], tokens_size)
            else:
                sim = self._calc_vector_similarity(((candidate_id + self.data.dataset_limit) if self.reverse_order else candidate_id), entity_id)
            self._insert_to_graph(
                candidate_id + self.data.dataset_limit if self.reverse_order else candidate_id,
                entity_id,
                sim
            )

    def _configuration(self) -> dict:
        return {
            "similarity_threshold" : self.similarity_threshold,
            "K" : self.K,
            "metric" : self.metric,
            "tokenization" : self.tokenization,
            "qgrams": self.qgrams
        }



import sys
import warnings
import pandas as pd
from itertools import chain
from collections import defaultdict
from logging import warning
from math import log10, sqrt
from queue import PriorityQueue
from time import time

import numpy as np
import math
from tqdm.autonotebook import tqdm

from .evaluation import Evaluation

from .datamodel import Data, PYJEDAIFeature
from .utils import chi_square, create_entity_index, get_sorted_blocks_shuffled_entities, PositionIndex, canonical_swap, sorted_enumerate

from abc import ABC, abstractmethod
from typing import Tuple, List


class AbstractComparisonCleaning(PYJEDAIFeature):
    """Abstract class for Block cleaning
    """

    def __init__(self) -> None:
        super().__init__()
        self._limit: int
        self._num_of_blocks: int
        self._valid_entities: set() = set()
        self._entity_index: dict
        self._weighting_scheme: str
        self._blocks: dict # initial blocks
        self.blocks = dict() # blocks after CC
        self._node_centric: bool

    def process(
            self,
            blocks: dict,
            data: Data,
            tqdm_disable: bool = False,
            store_weights: bool = False
    ) -> dict:
        """Main method for comparison cleaning

        Args:
            blocks (dict): blocks creted from previous steps of pyjedai
            data (Data): dataset module
            tqdm_disable (bool, optional): Disables tqdm progress bars. Defaults to False.

        Returns:
            dict: cleaned blocks
        """
        start_time = time()
        self.tqdm_disable, self.data, self.store_weights = tqdm_disable, data, store_weights
        self._limit = self.data.num_of_entities \
                if self.data.is_dirty_er or self._node_centric else self.data.dataset_limit
        self._progress_bar = tqdm(
            total=self._limit,
            desc=self._method_name,
            disable=self.tqdm_disable
        )

        self._stored_weights =  defaultdict(float) if self.store_weights else None
        self._entity_index = create_entity_index(blocks, self.data.is_dirty_er)
        self._num_of_blocks = len(blocks)
        self._blocks: dict = blocks
        self._blocks = self._apply_main_processing()
        self.execution_time = time() - start_time
        self._progress_bar.close()

        return self._blocks

    def report(self) -> None:
        """Prints Block Building method configuration
        """
        print(
            "Method name: " + self._method_name +
            "\nMethod info: " + self._method_info +
            ("\nParameters: \n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in self._configuration().items()]) if self._configuration().items() else "\nParameters: Parameter-Free method\n") +
            "Attributes from D1:\n\t" + ', '.join(c for c in (self.attributes_1 if self.attributes_1 is not None \
                else self.data.dataset_1.columns)) +
            ("\nAttributes from D2:\n\t" + ', '.join(c for c in (self.attributes_2 if self.attributes_2 is not None \
                else self.data.dataset_2.columns)) if not self.data.is_dirty_er else "") +
            "\nRuntime: {:2.4f} seconds".format(self.execution_time)
        )

    @abstractmethod
    def _apply_main_processing(self) -> dict:
        pass

    @abstractmethod
    def _configuration(self) -> dict:
        pass
    
    def stats(self) -> None:
        pass
    
    def get_precalculated_weight(self, entity_id: int, neighbor_id: int) -> float:
        """Returns the precalculated weight for given pair

        Args:
            entity_id (int): Entity ID
            neighbor_id (int): Neighbor ID

        Raises:
            AttributeError: Given pair has no precalculated weigth

        Returns:
            float: Pair weigth
        """
        if(not self.store_weights): raise AttributeError("No precalculated weights.")
        _weight = self._stored_weights.get(canonical_swap(entity_id, neighbor_id), KeyError(f"Pair [{entity_id},{neighbor_id}] has no precalculated weight"))
        return _weight
        
    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:

        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " +
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        total_matching_pairs = sum([len(block) for block in prediction.values()])
        for _, (id1, id2) in self.data.ground_truth.iterrows():
            id1 = self.data._ids_mapping_1[id1]
            id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er \
                                                else self.data._ids_mapping_2[id2]
            if (id1 in prediction and id2 in prediction[id1]) or   \
                (id2 in prediction and id1 in prediction[id2]):
                true_positives += 1

        eval_obj.calculate_scores(true_positives=true_positives, 
                                  total_matching_pairs=total_matching_pairs)
        return eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)

    def export_to_df(self, prediction) -> pd.DataFrame:
        """creates a dataframe with the predicted pairs

        Args:
            prediction (any): Predicted candidate pairs

        Returns:
            pd.DataFrame: Dataframe with the predicted pairs
        """
        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. \
                Data object mush have initialized with the ground-truth file")
        pairs_df = pd.DataFrame(columns=['id1', 'id2'])
        
        for entity_id, candidates in prediction.items():
            id1 = self.data._gt_to_ids_reversed_1[entity_id]                                            
            for candiadate_id in candidates:
                id2 = self.data._gt_to_ids_reversed_1[candiadate_id] if self.data.is_dirty_er \
                        else self.data._gt_to_ids_reversed_2[candiadate_id]
                pairs_df = pd.concat([pairs_df, pd.DataFrame([{'id1':id1, 'id2':id2}], index=[0])], ignore_index=True)

        return pairs_df


class AbstractMetablocking(AbstractComparisonCleaning, ABC):
    """Restructure a redundancy-positive block collection into a new
        one that contains substantially lower number of redundant
        and superfluous comparisons, while maintaining the original number of matching ones
    """

    def __init__(self) -> None:
        super().__init__()
        self._flags: np.array
        self._counters: np.array
        self._flags: np.array
        self._comparisons_per_entity: np.array
        self._node_centric: bool
        self._threshold: float
        self._distinct_comparisons: int
        self._comparisons_per_entity: np.array
        self._neighbors: set() = set()
        self._retained_neighbors: set() = set()
        self._block_assignments: int = 0
        self.weighting_scheme: str

    def _apply_main_processing(self) -> dict:
        self._counters = np.empty([self.data.num_of_entities], dtype=float)
        self._flags = np.empty([self.data.num_of_entities], dtype=int)
        if(self._comparisons_per_entity_required()):
            self._set_statistics()
        self._set_threshold()

        return self._prune_edges()

    def _comparisons_per_entity_required(self):
        return (self.weighting_scheme == 'EJS' or 
                self.weighting_scheme == 'CNC' or
                self.weighting_scheme == 'SNC' or
                self.weighting_scheme == 'SND' or
                self.weighting_scheme == 'CND' or
                self.weighting_scheme == 'CNJ' or
                self.weighting_scheme == 'SNJ')
        
    def _get_weight(self, entity_id: int, neighbor_id: int) -> float:
        ws = self.weighting_scheme
        if ws == 'CN-CBS' or ws == 'CBS' or ws == 'SN-CBS':
            return self._counters[neighbor_id]
        # CARDINALITY_NORM_COSINE, SIZE_NORM_COSINE
        elif ws == 'CNC' or ws == 'SNC':
            return self._counters[neighbor_id] / float(sqrt(self._comparisons_per_entity[entity_id] * self._comparisons_per_entity[neighbor_id]))
        # SIZE_NORM_DICE, CARDINALITY_NORM_DICE
        elif ws == 'SND' or ws == 'CND':
            return 2 * self._counters[neighbor_id] / float(self._comparisons_per_entity[entity_id] + self._comparisons_per_entity[neighbor_id])
        # CARDINALITY_NORM_JS, SIZE_NORM_JS
        elif ws == 'CNJ' or ws == 'SNJ':
            return  self._counters[neighbor_id] / float(self._comparisons_per_entity[entity_id] + self._comparisons_per_entity[neighbor_id] - self._counters[neighbor_id])
        elif ws == 'COSINE':
            return self._counters[neighbor_id] / float(sqrt(len(self._entity_index[entity_id]) * len(self._entity_index[neighbor_id])))
        elif ws == 'DICE':
            return 2 * self._counters[neighbor_id] / float(len(self._entity_index[entity_id]) + len(self._entity_index[neighbor_id]))
        elif ws == 'ECBS':
            return float(
                self._counters[neighbor_id] *
                log10(float(self._num_of_blocks / len(self._entity_index[entity_id]))) *
                log10(float(self._num_of_blocks / len(self._entity_index[neighbor_id])))
            )
        elif ws == 'JS':
            return self._counters[neighbor_id] / (len(self._entity_index[entity_id]) + \
                    len(self._entity_index[neighbor_id]) - self._counters[neighbor_id])
        elif ws == 'EJS':
            probability = self._counters[neighbor_id] / (len(self._entity_index[entity_id]) + \
                            len(self._entity_index[neighbor_id]) - self._counters[neighbor_id])
            return float(probability * \
                    log10(self._distinct_comparisons / self._comparisons_per_entity[entity_id]) * \
                    log10(self._distinct_comparisons / self._comparisons_per_entity[neighbor_id]))
        elif ws == 'X2':
            observed = [int(self._counters[neighbor_id]),
                        int(len(self._entity_index[entity_id])-self._counters[neighbor_id])]
            expected = [int(len(self._entity_index[neighbor_id])-observed[0]),
                        int(self._num_of_blocks - (observed[0] + observed[1] - self._counters[neighbor_id]))]
            return chi_square(np.array([observed, expected]))
        else:
            raise ValueError("This weighting scheme does not exist")

    def _normalize_neighbor_entities(self, block_key: str, entity_id: int) -> None:
        self._neighbors.clear()
        if self.data.is_dirty_er:
            if not self._node_centric:
                self._neighbors.update(
                    [neighbor_id for neighbor_id in self._blocks[block_key].entities_D1 \
                        if neighbor_id < entity_id]
                )
            else:
                self._neighbors.update(
                    [neighbor_id for neighbor_id in self._blocks[block_key].entities_D1 \
                        if neighbor_id != entity_id]
                )
        else:
            if entity_id < self.data.dataset_limit:
                self._neighbors.update(self._blocks[block_key].entities_D2)
            else:
                self._neighbors.update(self._blocks[block_key].entities_D1)

    def _set_statistics(self) -> None:
        self._distinct_comparisons = 0
        self._comparisons_per_entity = np.empty([self.data.num_of_entities], dtype=float)
        distinct_neighbors = set()
        for entity_id in range(0, self.data.num_of_entities, 1):
            if entity_id in self._entity_index:
                associated_blocks = self._entity_index[entity_id]
                distinct_neighbors.clear()
                # distinct_neighbors = set().union(*[
                #     self._get_neighbor_entities(block_id, entity_id) for block_id in associated_blocks
                # ])
                distinct_neighbors = set(chain.from_iterable(
                    self._get_neighbor_entities(block_id, entity_id) for block_id in associated_blocks
                ))
                # for block_id in associated_blocks:
                #     distinct_neighbors = set.union(
                #         distinct_neighbors,
                #         self._get_neighbor_entities(block_id, entity_id)
                #     )
                self._comparisons_per_entity[entity_id] = len(distinct_neighbors)

                if self.data.is_dirty_er:
                    self._comparisons_per_entity[entity_id] -= 1

                self._distinct_comparisons += self._comparisons_per_entity[entity_id]
        self._distinct_comparisons /= 2

    def _get_neighbor_entities(self, block_id: int, entity_id: int) -> set:
        return self._blocks[block_id].entities_D2 \
            if (not self.data.is_dirty_er and entity_id < self.data.dataset_limit) else \
                self._blocks[block_id].entities_D1

    @abstractmethod
    def _set_threshold(self):
        pass

    @abstractmethod
    def _prune_edges(self) -> dict:
        pass

class ComparisonPropagation(AbstractComparisonCleaning):
    """Comparison Propagation
    """

    _method_name = "Comparison Propagation"
    _method_short_name: str = "CP"
    _method_info = "Eliminates all redundant comparisons from a set of overlapping blocks."

    def __init__(self) -> None:
        super().__init__()
        self._node_centric = False

    def _apply_main_processing(self) -> dict:
        self.blocks = {}
        for i in range(0, self._limit):
            if i in self._entity_index:
                self._valid_entities.clear()
                associated_blocks = self._entity_index[i]
                for block_index in associated_blocks:
                    if self.data.is_dirty_er:
                        self._valid_entities.update(
                            [neighbor_id for neighbor_id in self._blocks[block_index].entities_D1 if i < neighbor_id]
                        )
                    else:
                        self._valid_entities.update(self._blocks[block_index].entities_D2)
                self.blocks[i] = self._valid_entities.copy()
            self._progress_bar.update(1)
        return self.blocks
 
    def _configuration(self) -> dict:
        return {
            "Node centric" : self._node_centric
        }

class WeightedEdgePruning(AbstractMetablocking):
    """A Meta-blocking method that retains all comparisons \
        that have a weight higher than the average edge weight in the blocking graph.
    """

    _method_name = "Weighted Edge Pruning"
    _method_short_name: str = "WEP"
    _method_info = "A Meta-blocking method that retains all comparisons " + \
                "that have a weight higher than the average edge weight in the blocking graph."

    def __init__(self, weighting_scheme: str = 'CBS') -> None:
        super().__init__()
        self.weighting_scheme = weighting_scheme
        self._node_centric = False
        self._num_of_edges: float

    def _prune_edges(self) -> dict:
        for i in range(0, self._limit):
            self._process_entity(i)
            self._verify_valid_entities(i)
            self._progress_bar.update(1)

        return self.blocks

    def _process_entity(self, entity_id: int) -> None:
        if entity_id not in self._entity_index:
            return
        self._valid_entities.clear()
        self._flags[:] = -1
        associated_blocks = self._entity_index[entity_id]
        for block_id in associated_blocks:
            self._normalize_neighbor_entities(block_id, entity_id)
            for neighbor_id in self._neighbors:
                if self._flags[neighbor_id] != entity_id:
                    self._counters[neighbor_id] = 0
                    self._flags[neighbor_id] = entity_id
                if self.weighting_scheme == 'CN-CBS' or self.weighting_scheme == 'CNC' or self.weighting_scheme == 'CND' or self.weighting_scheme == 'CNJ':
                    self._counters[neighbor_id] += 1 / self._blocks[block_id].get_cardinality(self.data.is_dirty_er)
                if self.weighting_scheme == 'SN-CBS' or self.weighting_scheme == 'SNC' or self.weighting_scheme == 'SND' or self.weighting_scheme == 'SNJ':
                    self._counters[neighbor_id] += 1 / self._blocks[block_id].get_size()
                else:
                    self._counters[neighbor_id] += 1
                self._valid_entities.add(neighbor_id)

    def _update_threshold(self, entity_id: int) -> None:
        if entity_id not in self._entity_index:
            return

        self._num_of_edges += len(self._valid_entities)
        for neighbor_id in self._valid_entities:
            self._threshold += self._get_weight(entity_id, neighbor_id)

    def _set_threshold(self):
        self._num_of_edges = 0.0
        self._threshold = 0.0
        for i in range(0, self._limit):
            self._process_entity(i)
            self._update_threshold(i)

        self._threshold /= self._num_of_edges

    def _verify_valid_entities(self, entity_id: int) -> None:    
        if entity_id not in self._entity_index:
            return

        self._retained_neighbors.clear()
        for neighbor_id in self._valid_entities:
            weight = self._get_weight(entity_id, neighbor_id)
            if self._threshold <= weight:
                self._retained_neighbors.add(neighbor_id)
                if self.store_weights:
                    self._stored_weights[canonical_swap(entity_id, neighbor_id)] = weight

        if len(self._retained_neighbors) > 0:
            self.blocks[entity_id] = self._retained_neighbors.copy()

    def _configuration(self) -> dict:
        return {
            "Node centric" : self._node_centric,
            "Weighting scheme" : self.weighting_scheme
        }

class CardinalityEdgePruning(WeightedEdgePruning):
    """A Meta-blocking method that retains the comparisons \
            that correspond to the top-K weighted edges in the blocking graph.
    """

    _method_name = "Cardinality Edge Pruning"
    _method_short_name: str = "CEP"
    _method_info = "A Meta-blocking method that retains the comparisons " + \
                        "that correspond to the top-K weighted edges in the blocking graph."

    def __init__(self, weighting_scheme: str = 'JS') -> None:
        super().__init__(weighting_scheme)
        self._minimum_weight: float = sys.float_info.min
        self._top_k_edges: PriorityQueue

    def _prune_edges(self) -> dict:
        self.blocks = defaultdict(set)
        self._top_k_edges = PriorityQueue(int(2*self._threshold))
        for i in range(0, self._limit):
            self._process_entity(i)
            self._verify_valid_entities(i)
            self._progress_bar.update(1)
        while not self._top_k_edges.empty():
            comparison = self._top_k_edges.get()
            self.blocks[comparison[1]].add(comparison[2])
            if self.store_weights:
                self._stored_weights[canonical_swap(comparison[1], comparison[2])] = comparison[0]

        return self.blocks

    def _set_threshold(self) -> None:
        block_assignments = 0
        for block in self._blocks.values():
            block_assignments += block.get_size()
        self._threshold = block_assignments / 2

    def _verify_valid_entities(self, entity_id: int) -> None:
        if entity_id not in self._entity_index:
            return

        for neighbor_id in self._valid_entities:
            weight = self._get_weight(entity_id, neighbor_id)
            if weight >= self._minimum_weight:
                self._top_k_edges.put(
                    (weight, entity_id, neighbor_id)
                )
                if self._threshold < self._top_k_edges.qsize():
                    self._minimum_weight = self._top_k_edges.get()[0]

class CardinalityNodePruning(CardinalityEdgePruning):
    """A Meta-blocking method that retains for every entity, \
        the comparisons that correspond to its top-k weighted edges in the blocking graph."
    """

    _method_name = "Cardinality Node Pruning"
    _method_short_name: str = "CNP"
    _method_info = "A Meta-blocking method that retains for every entity, " + \
                    "the comparisons that correspond to its top-k weighted edges in the blocking graph."

    def __init__(self, weighting_scheme: str = 'CBS') -> None:
        super().__init__(weighting_scheme)
        self._nearest_entities: dict
        self._node_centric = True
        self._top_k_edges: PriorityQueue
        self._number_of_nearest_neighbors : int = None

    def _prune_edges(self) -> dict:
        self._nearest_entities = dict()
        for i in range(0, self._limit):
            self._process_entity(i)
            self._verify_valid_entities(i)
            self._progress_bar.update(1)
        return self._retain_valid_comparisons()

    def _retain_valid_comparisons(self) -> dict:
        self.blocks = dict()
        for i in range(0, self.data.num_of_entities):
            if i in self._nearest_entities:
                for neighbor_id in self._nearest_entities[i]:
                    # Comparison is a triple: (id1, id2, weight)
                    if self._is_valid_comparison(i, neighbor_id):
                        self.blocks.setdefault(i, set())
                        self.blocks[i].add(neighbor_id)
        return self.blocks

    def _is_valid_comparison(self, entity_id: int, neighbor_id: int) -> bool:
        if neighbor_id not in self._nearest_entities:
            return True
        if entity_id in self._nearest_entities[neighbor_id]:
            return entity_id < neighbor_id
        return True

    def _set_threshold(self) -> None:
        if(self._number_of_nearest_neighbors is None):
            block_assignments = 0
            for block in self._blocks.values():
                block_assignments += block.get_size()
            self._threshold = max(1, block_assignments / self.data.num_of_entities)
        else:
            self._threshold = self._number_of_nearest_neighbors         

    def _verify_valid_entities(self, entity_id: int) -> None:
        if entity_id not in self._entity_index:
            return
        self._top_k_edges = PriorityQueue(int(2*self._threshold))
        self._minimum_weight = sys.float_info.min
        for neighbor_id in self._valid_entities:
            weight = self._get_weight(entity_id, neighbor_id)
            if weight >= self._minimum_weight:
                self._top_k_edges.put(
                    (weight, entity_id, neighbor_id)
                )
                if self._threshold < self._top_k_edges.qsize():
                    self._minimum_weight = self._top_k_edges.get()[0]
        if self._top_k_edges:         
            self._nearest_entities.setdefault(entity_id, set())
        while not self._top_k_edges.empty():
            comparison = self._top_k_edges.get()
            self._nearest_entities[entity_id].add(comparison[2])
            if self.store_weights:
                    self._stored_weights[canonical_swap(entity_id, comparison[2])] = comparison[0]

class ReciprocalCardinalityNodePruning(CardinalityNodePruning):
    """A Meta-blocking method that retains the comparisons \
        that correspond to edges in the blocking graph that are among the top-k weighted  \
        ones for both adjacent entities/nodes.
    """

    _method_name = "Reciprocal Cardinality Node Pruning"
    _method_short_name: str = "RCNP"
    _method_info = "A Meta-blocking method that retains the comparisons " + \
                    "that correspond to edges in the blocking graph that are among " + \
                    "the top-k weighted ones for both adjacent entities/nodes."

    def __init__(self, weighting_scheme: str = 'CN-CBS') -> None:
        super().__init__(weighting_scheme)

    def _is_valid_comparison(self, entity_id: int, neighbor_id: int) -> bool:
        if neighbor_id not in self._nearest_entities:
            return False
        if entity_id in self._nearest_entities[neighbor_id]:
            return  entity_id < neighbor_id
        return False

class WeightedNodePruning(WeightedEdgePruning):
    """A Meta-blocking method that retains for every entity, the comparisons \
        that correspond to edges in the blocking graph that are exceed \
        the average edge weight in the respective node neighborhood.
    """

    _method_name = "Weighted Node Pruning"
    _method_short_name: str = "WNP"
    _method_info = "A Meta-blocking method that retains for every entity, the comparisons \
                    that correspond to edges in the blocking graph that are exceed \
                    the average edge weight in the respective node neighborhood."

    def __init__(self, weighting_scheme: str = 'CBS') -> None:
        super().__init__(weighting_scheme)
        self._average_weight: np.array
        self._node_centric = True

    def _get_valid_weight(self, entity_id: int, neighbor_id: int) -> float:
        weight = self._get_weight(entity_id, neighbor_id)
        return weight if ((self._average_weight[entity_id] <= weight or \
                             self._average_weight[neighbor_id] <= weight) and 
                                entity_id < neighbor_id) else 0

    def _set_threshold(self):
        self._average_weight = np.zeros(self._limit, dtype=float)
        for i in range(0, self._limit):
            self._process_entity(i)
            self._update_threshold(i)

    def _update_threshold(self, entity_id: int) -> None:
        if entity_id not in self._entity_index:
            return
        self._average_weight[entity_id] = 0.0
        for neighbor_id in self._valid_entities:
            self._average_weight[entity_id] += super()._get_weight(entity_id, neighbor_id)
        self._average_weight[entity_id] /= len(self._valid_entities)

    def _verify_valid_entities(self, entity_id: int) -> None:
        if entity_id not in self._entity_index:
            return
        self._retained_neighbors.clear()
        for neighbor_id in self._valid_entities:
            _weight = self._get_valid_weight(entity_id, neighbor_id)
            if _weight:
                self._retained_neighbors.add(neighbor_id)
                if self.store_weights:
                    self._stored_weights[canonical_swap(entity_id, neighbor_id)] = _weight
        if len(self._retained_neighbors) > 0:
            self.blocks[entity_id] = self._retained_neighbors.copy()

class BLAST(WeightedNodePruning):
    """Meta-blocking method that retains the comparisons \
        that correspond to edges in the blocking graph that are exceed 1/4 of the sum \
        of the maximum edge weights in the two adjacent node neighborhoods.
    """

    _method_name = _method_short_name = "BLAST"
    _method_info = "Meta-blocking method that retains the comparisons " + \
                "that correspond to edges in the blocking graph that are exceed 1/4 of the sum " + \
                "of the maximum edge weights in the two adjacent node neighborhoods."

    def __init__(self, weighting_scheme: str = 'X2') -> None:
        super().__init__(weighting_scheme)

    def _get_valid_weight(self, entity_id: int, neighbor_id: int) -> float:
        weight = self._get_weight(entity_id, neighbor_id)
        edge_threshold = (self._average_weight[entity_id] + self._average_weight[neighbor_id]) / 4
        return edge_threshold <= weight and entity_id < neighbor_id

    def _update_threshold(self, entity_id: int) -> None:
        if entity_id not in self._entity_index:
            return
        self._average_weight[entity_id] = 0.0
        for neighbor_id in self._valid_entities:
            self._average_weight[entity_id] = \
                max(self._average_weight[entity_id], self._get_weight(entity_id, neighbor_id))

class ReciprocalWeightedNodePruning(WeightedNodePruning):
    """Meta-blocking method that retains the comparisons\
        that correspond to edges in the blocking graph that are \
        exceed the average edge weight in both adjacent node neighborhoods.
    """

    _method_name = "Reciprocal Weighted Node Pruning"
    _method_short_name: str = "RWNP"
    _method_info = "Meta-blocking method that retains the comparisons " + \
                    "that correspond to edges in the blocking graph that are " + \
                    "exceed the average edge weight in both adjacent node neighborhoods."

    def __init__(self, weighting_scheme: str = 'CN-CBS') -> None:
        super().__init__(weighting_scheme)

    def _get_valid_weight(self, entity_id: int, neighbor_id: int) -> float:
        weight = self._get_weight(entity_id, neighbor_id)
        return weight if ((self._average_weight[entity_id] <= weight and \
                             self._average_weight[neighbor_id] <= weight) and
                                entity_id < neighbor_id) else 0

class ProgressiveCardinalityEdgePruning(CardinalityEdgePruning):
    def __init__(self, weighting_scheme: str = 'JS', budget: int = 0) -> None:
        super().__init__(weighting_scheme)
        self._budget = budget

    def _set_threshold(self) -> None:
        self._threshold = self._budget

    def process(self, blocks: dict, data: Data, tqdm_disable: bool = False, store_weights: bool = True, cc: AbstractMetablocking = None, emit_all_tps_stop : bool = False) -> dict:
        
        self._emit_all_tps_stop : bool = emit_all_tps_stop
        self._budget = self._budget if not self._emit_all_tps_stop else float('inf')
        if(cc is None):
            return super().process(blocks, data, tqdm_disable, store_weights)
        else:
            self._threshold = self._budget
            self._top_k_edges = PriorityQueue(int(2*self._threshold))
            self._minimum_weight = sys.float_info.min            
            self.trimmed_blocks : dict = defaultdict(set)

            for entity_id, neighbors in blocks.items():
                for neighbor_id in neighbors:
                    weight = cc.get_precalculated_weight(entity_id, neighbor_id)
                    if weight >= self._minimum_weight:
                        self._top_k_edges.put(
                        (weight, entity_id, neighbor_id)
                        )
                        if self._threshold < self._top_k_edges.qsize():
                            self._minimum_weight = self._top_k_edges.get()[0]

            while not self._top_k_edges.empty():
                comparison = self._top_k_edges.get()
                self.trimmed_blocks[comparison[1]].add(comparison[2])
                if(self.store_weights):
                    self._stored_weights[canonical_swap(comparison[1], comparison[2])] = comparison[0]

            return self.trimmed_blocks

class ProgressiveCardinalityNodePruning(CardinalityNodePruning):
    def __init__(self, weighting_scheme: str = 'CBS', budget: int = 0) -> None:
        super().__init__(weighting_scheme)
        self._budget = budget

    def _set_threshold(self) -> None:
        self._threshold = self._number_of_nearest_neighbors

    def process(self, blocks: dict,
                data: Data,
                number_of_nearest_neighbors : int = 10,
                tqdm_disable: bool = False,
                store_weights: bool = True,
                cc: AbstractMetablocking = None,
                emit_all_tps_stop : bool = False) -> dict:
        self._emit_all_tps_stop : bool = emit_all_tps_stop
        self._number_of_nearest_neighbors : int = number_of_nearest_neighbors
        if(cc is None):
            return super().process(blocks=blocks, data=data, tqdm_disable=tqdm_disable, store_weights=store_weights)
            
        else:
            self._threshold = self._number_of_nearest_neighbors         
            self.trimmed_blocks : dict = defaultdict(set)

            for entity_id, neighbors in blocks.items():
                self._minimum_weight = sys.float_info.min
                self._top_k_edges = PriorityQueue(int(2*self._threshold))
                for neighbor_id in neighbors:
                    weight = cc.get_precalculated_weight(entity_id, neighbor_id)
                    if weight >= self._minimum_weight:
                        self._top_k_edges.put(
                        (weight, entity_id, neighbor_id)
                        )
                        if self._threshold < self._top_k_edges.qsize():
                            self._minimum_weight = self._top_k_edges.get()[0]

                while not self._top_k_edges.empty():
                    comparison = self._top_k_edges.get()
                    self.trimmed_blocks[entity_id].add(comparison[2])
                    if self.store_weights:
                        self._stored_weights[canonical_swap(entity_id, comparison[2])] = comparison[0]

        return self.trimmed_blocks 
    
    
class ProgressiveSortedNeighborhood(AbstractMetablocking):
    """Progressive Sorted Neighborhood"""

    _method_name = "Progressive Sorted Neighborhood"
    _method_short_name: str = "PSN"
    _method_info = "Sorts and iterates over entities in an incremental, windowed manner, compares the entities within defined windows " + \
                    "and orders non-reduntant comparisons within the windows by decreasing frequency"

    def __init__(self, weighting_scheme: str = 'ACF', budget: int = 0) -> None:
        self.weighting_scheme: str = weighting_scheme
        self._budget : int = budget
        super().__init__()
        self._node_centric = False
        
    
    def process(
            self,
            blocks: dict,
            data: Data,
            window_size : int = 10,
            tqdm_disable: bool = False,
            emit_all_tps_stop : bool = False
    ) -> List[Tuple[float, int, int]]:
        """Calculates top comparisons for Progressive Matching

        Args:
            blocks (dict): blocks creted from previous steps of pyjedai
            data (Data): dataset module
            tqdm_disable (bool, optional): Disables tqdm progress bars. Defaults to False.

        Returns:
            PriorityQueue: Top Comparisons
        """
        start_time = time()
        self.tqdm_disable, self.data = tqdm_disable, data
        self._limit = self.data.num_of_entities \
                if self.data.is_dirty_er or self._node_centric else self.data.dataset_limit
        self._progress_bar = tqdm(
            total=self._limit,
            desc=self._method_name,
            disable=self.tqdm_disable
        )
        self._emit_all_tps_stop : bool = emit_all_tps_stop
        self._num_of_blocks = len(blocks)
        self._blocks: dict = blocks
        self._max_window_size : int = window_size
        
        self._sorted_entity_ids = get_sorted_blocks_shuffled_entities(self.data.is_dirty_er, self._blocks)
        self._total_sorted_entities = len(self._sorted_entity_ids)
        self._position_index = PositionIndex(self.data.num_of_entities, self._sorted_entity_ids)
        
        self._counters = np.empty([self.data.num_of_entities], dtype=float)
        self._flags = np.empty([self.data.num_of_entities], dtype=int)
        self._counters[:] = 0
        self._flags[:] = -1
        self._pairs : List[Tuple[float, int, int]]= self._apply_main_processing()
        self.execution_time = time() - start_time
        self._progress_bar.close()

        return self._pairs    
    
    def _get_weight(self, entity_id: int, neighbor_id: int) -> float:
        ws = self.weighting_scheme
        
        if ws == 'NCF':
            denominator : float = len(self._position_index.get_positions(entity_id)) + len(self._position_index.get_positions(neighbor_id)) - self._counters[neighbor_id]
            return self._counters[neighbor_id] / denominator
        elif ws == 'ACF' or ws == 'ID':
            return self._counters[neighbor_id]
        elif ws == 'COSINE':
            return self._counters[neighbor_id] / float(sqrt(len(self._position_index.get_positions(entity_id)) * len(self._position_index.get_positions(neighbor_id))))
        elif ws == 'DICE':
            return 2 * self._counters[neighbor_id] / float(len(self._position_index.get_positions(entity_id)) + len(self._position_index.get_positions(neighbor_id)))
        else:
            raise ValueError("This weighting scheme does not exist")    
        
    def valid_entity_neighbor_index(self, entity: int, neighbor_index: int) -> bool:
        """Verifies if the neighbor identifier at the specified index is valid for candidate (the pair hasn't been considered previously)

        Args:
            entity (int): Identifier of the current entity
            neighbor_index (int): Index of the neighbor identifier within the list of sorted indices

        Returns:
            bool: Valid / Not Valid candidate for pair
        """
        neighbor = self._sorted_entity_ids[neighbor_index]
        return self.data.dataset_limit <= neighbor if not self.data.is_dirty_er else neighbor < entity
    
    def _set_threshold(self):
        pass
    
    def _prune_edges(self) -> dict:
        pass
    
    def _configuration(self) -> dict:
        pass        
        
class GlobalProgressiveSortedNeighborhood(ProgressiveSortedNeighborhood):
    """Global Progressive Sorted Neighborhood"""

    _method_name = "Global Progressive Sorted Neighborhood"
    _method_short_name: str = "GPSN"
    _method_info = "For each sorted entity, conducts incrementally expanding window wise iteration over all the sorted entities, " + \
                    "calculates local scores for the entities present within current window and stores the best comparisons on a global scale"
        
    def __init__(self, weighting_scheme: str = 'ACF', budget: int = 0) -> None:
        super().__init__(weighting_scheme, budget)
        
    def _apply_main_processing(self) -> List[Tuple[float, int, int]]:
        # TO DO: budget taken as argument in prediction, not algorithm constructor
        self._budget = float('inf') if self._emit_all_tps_stop else self._budget
        self._top_pairs : List[Tuple[float, int, int]] = []
        default_weight = 0.0
        self._pair_weight : dict = defaultdict(lambda: default_weight)
        
        for entity in range(self.data.dataset_limit):
            entity_positions = self._position_index.get_positions(entity)
            self._neighbors.clear()
            for current_window in range(1,self._max_window_size + 1):
                for entity_position in entity_positions:
                    right_neighbor = entity_position + current_window
                    left_neighbor = entity_position - current_window

                    if(right_neighbor < self._total_sorted_entities):
                         if(self.valid_entity_neighbor_index(entity, right_neighbor)):
                            self._update_local_weight(current_window, entity, self._sorted_entity_ids[right_neighbor])
                    if(left_neighbor >= 0):
                        if(self.valid_entity_neighbor_index(entity, left_neighbor)):
                            self._update_local_weight(current_window, entity, self._sorted_entity_ids[left_neighbor])
                      
            for neighbor in self._neighbors:
                self._flags[neighbor] = -1
                self._pair_weight[(entity, neighbor)] = max(self._pair_weight[(entity, neighbor)], self._get_weight(entity, neighbor))
                
        for pair in self._pair_weight:
            id1, id2 = pair
            self._top_pairs.append((self._pair_weight[(id1, id2)], id1, id2))
                        
        return self._top_pairs
                                            
    def _update_local_weight(self, window : int, entity: int, neighbor: int):
        """Updates the weight of the entity & neighbor pair for current window

        Args:
            window (int): Current Window Size
            entity (int): Current Entity ID
            neighbor (int): Current Neighbor ID
        """
        
        if(self._flags[neighbor] != entity):
            self._counters[neighbor] = 0
            self._flags[neighbor] = entity
        
        pwScheme = self.weighting_scheme
        if pwScheme == 'ID':
            self._counters[neighbor] += 1.0 / window
        else:
            self._counters[neighbor] += 1.0
            
        self._neighbors.add(neighbor)

class LocalProgressiveSortedNeighborhood(ProgressiveSortedNeighborhood):
    """Local Progressive Sorted Neighborhood"""

    _method_name = "Local Progressive Sorted Neighborhood"
    _method_short_name: str = "LPSN"
    _method_info = "Iteratively increments window size. For each one, derives the distinct neighbors for each entity, " + \
                    "calculates their similarity, and emits the pairs in decreasing similarity score order"
        
    def __init__(self, weighting_scheme: str = 'ACF', budget: int = 0) -> None:
        super().__init__(weighting_scheme, budget)
        
    def _has_next(self) -> bool:
        """Validates if more pairs can be emitted

        Returns:
            bool: Another pair can be emitted
        """
        return self._current_window <= self._max_window_size
        
    def _apply_main_processing(self) -> List[Tuple[float, int, int]]:
        self._current_window = 1 
        self._top_pairs: List[Tuple[float, int, int]] = []
        default_weight = 0.0
        self._pair_weight : dict = defaultdict(lambda: default_weight)
        # TO DO: budget taken as argument in prediction, not algorithm constructor
        self._budget = float('inf') if self._emit_all_tps_stop else self._budget
        
        while(self._has_next()):
            for entity in range(self.data.dataset_limit):
                entity_positions = self._position_index.get_positions(entity)
                self._neighbors.clear()
            
                for entity_position in entity_positions:
                    right_neighbor = entity_position + self._current_window
                    left_neighbor = entity_position - self._current_window
                    
                    if(right_neighbor < self._total_sorted_entities):
                        if(self.valid_entity_neighbor_index(entity, right_neighbor)):
                            self._update_counters(entity, self._sorted_entity_ids[right_neighbor])
                    if(left_neighbor >= 0):
                        if(self.valid_entity_neighbor_index(entity, left_neighbor)):
                            self._update_counters(entity, self._sorted_entity_ids[left_neighbor])
                 
                for neighbor in self._neighbors:
                    self._flags[neighbor] = -1
                    self._pair_weight[(entity, neighbor)] = max(self._pair_weight[(entity, neighbor)], self._get_weight(entity, neighbor))

            self._current_window += 1
            
        for pair in self._pair_weight:
            id1, id2 = pair
            self._top_pairs.append((self._pair_weight[(id1, id2)], id1, id2))
           
        return self._top_pairs
                                            
    def _update_counters(self, entity: int, neighbor: int):
        """Updates the counters of the entity & neighbor pair for current window

        Args:
            entity (int): Current Entity ID
            neighbor (int): Current Neighbor ID
        """
        
        if(self._flags[neighbor] != entity):
            self._counters[neighbor] = 0
            self._flags[neighbor] = entity
        
        self._counters[neighbor] += 1.0    
        self._neighbors.add(neighbor)
        
        
class ProgressiveEntityScheduling(WeightedNodePruning):
    """Progressive Entity Scheduling"""

    _method_name = "Progressive Entity Scheduling"
    _method_short_name: str = "PES"
    _method_info = "Sorts entities in descending order of their average weight, " + \
                    "emits the top pair per entity. Finally, traverses the sorted " + \
                    "entities and emits their comparisons in descending weight order " + \
                    "within specified budget"
    def __init__(self, weighting_scheme: str = 'CBS', budget: int = 0) -> None:
        super().__init__(weighting_scheme)
        self._budget = budget

    def _process_entity(self, entity_id: int) -> None:
        """Calculates the counters for the neighbors of specified entity,
           stores the weight for each neighbor and the top comparison for current entity.
           Finally, creates a prunned block for specified entity

        Args:
            entity_id (int): Entity ID
        """
        if entity_id not in self._entity_index:
            self.blocks[entity_id] = set()
            return
        self._valid_entities.clear()
        self._flags[:] = -1
        associated_blocks = self._entity_index[entity_id]
        
        for block_id in associated_blocks:
            self._normalize_neighbor_entities(block_id, entity_id)
            for neighbor_id in self._neighbors:
                if self._flags[neighbor_id] != entity_id:
                    self._counters[neighbor_id] = 0
                    self._flags[neighbor_id] = entity_id
                if self.weighting_scheme == 'CN-CBS' or self.weighting_scheme == 'CNC' or self.weighting_scheme == 'CND' or self.weighting_scheme == 'CNJ':
                    self._counters[neighbor_id] += 1 / self._blocks[block_id].get_cardinality(self.data.is_dirty_er)
                if self.weighting_scheme == 'SN-CBS' or self.weighting_scheme == 'SNC' or self.weighting_scheme == 'SND' or self.weighting_scheme == 'SNJ':
                    self._counters[neighbor_id] += 1 / self._blocks[block_id].get_size()
                else:
                    self._counters[neighbor_id] += 1
                self._valid_entities.add(neighbor_id)
                       
        for valid_entity_id in self._valid_entities:  
            _current_neighbor_weight = self._get_weight(entity_id, valid_entity_id)
            if(self.store_weights):
                self._stored_weights[canonical_swap(entity_id, valid_entity_id)] = _current_neighbor_weight
                
            self._to_emit_pairs.append((_current_neighbor_weight, entity_id, valid_entity_id))
        self.blocks[entity_id] = self._valid_entities.copy()        

    def _prune_edges(self) -> dict:
        return None

    def process_raw_blocks(self, blocks: dict):
        self._average_weight = np.zeros(self._limit, dtype=float)
        self._entity_index = create_entity_index(blocks, self.data.is_dirty_er)
        self._apply_main_processing()
        
    def process_prunned_blocks(self, blocks : dict, cc : AbstractMetablocking):
        self.blocks = blocks
        for entity in sorted(blocks.keys()):
            neighbors = blocks[entity]
            for neighbor in neighbors:
                _current_neighbor_weigth = cc.get_precalculated_weight(entity, neighbor) 
                self._to_emit_pairs.append((_current_neighbor_weigth, entity, neighbor))
    

    def process(self, blocks: dict, data: Data, tqdm_disable: bool = False, store_weigths : bool = True, cc: AbstractMetablocking = None, method : str = 'HB', emit_all_tps_stop : bool = False) -> List[Tuple[float, int, int]]:
        """Calculates the weights between entities, stores them in descending order of their average weight,
           stores the top comparison per entity

        Args:
            blocks (dict): Blocks to process
            data (Data): Data Feature
            tqdm_disable (bool, optional): Progress Bar. Defaults to False.
            cc (AbstractMetablocking, optional): Comparison Cleaner used in previous step. Defaults to None.

        Returns:
            None: None
        """
        
        self.start_time = time()
        self.tqdm_disable, self.data, self.store_weights, self.method = tqdm_disable, data, store_weigths, method
        self._limit = self.data.num_of_entities \
                if self.data.is_dirty_er or self._node_centric else self.data.dataset_limit
        self._progress_bar = tqdm(
            total=self._limit,
            desc=self._method_name,
            disable=self.tqdm_disable
        )
        
        self._emit_all_tps_stop : bool = emit_all_tps_stop
        self._num_of_blocks = len(blocks)
        self._blocks: dict = blocks
        self._stored_weights : dict = defaultdict(float)
        self._to_emit_pairs = []

        if(cc is None):
            self.process_raw_blocks(blocks)
        else:
            self.process_prunned_blocks(blocks, cc)
            
        return self._to_emit_pairs
            
def get_meta_blocking_approach(acronym: str, w_scheme: str, budget: int = 0) -> any:
    """Return method by acronym

    Args:
        acronym (str): Method acronym
        w_scheme (str): Weighting Scheme name

    Returns:
        any: Comparison Cleaning Method
    """
    if acronym == "BLAST":
        return BLAST(w_scheme)
    elif acronym == "CEP":
        return CardinalityEdgePruning(w_scheme)
    elif acronym == "CNP":
        return CardinalityNodePruning(w_scheme)
    elif acronym == "RCNP":
        return ReciprocalCardinalityNodePruning(w_scheme)
    elif acronym == "RWNP":
        return ReciprocalWeightedNodePruning(w_scheme)
    elif acronym == "WEP":
        return WeightedEdgePruning(w_scheme)
    elif acronym == "WNP":
        return WeightedNodePruning(w_scheme)
    elif acronym == "PCEP":
        return ProgressiveCardinalityEdgePruning(w_scheme, budget)
    elif acronym == "PCNP":
        return ProgressiveCardinalityNodePruning(w_scheme, budget)
    elif acronym == "GPSN":
        return GlobalProgressiveSortedNeighborhood(w_scheme, budget)
    elif acronym == "LPSN":
        return LocalProgressiveSortedNeighborhood(w_scheme, budget)
    elif acronym == "PES":
        return ProgressiveEntityScheduling(w_scheme, budget)
    else:
        warnings.warn("Wrong meta-blocking approach selected. Returning Comparison Propagation.")
        return ComparisonPropagation()



import logging
import sys
from logging import WARN  # noqa

_interactive = False
try:
    # This is only defined in interactive shells
    if sys.ps1:
        _interactive = True
except AttributeError:
    # Even now, we may be in an interactive shell with `python -i`.
    _interactive = sys.flags.interactive

_logger = logging.getLogger('pyjedai')

if _interactive:
    _logger.setLevel(WARN)
    _logging_target = sys.stdout
else:
    _logging_target = sys.stderr

_handler = logging.StreamHandler(sys.stdout)
_handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT, None))
_logger.addHandler(_handler)

log = _logger.log
debug = _logger.debug
error = _logger.error
fatal = _logger.fatal
info = _logger.info
warning = _logger.warning



"""Block Cleaning Module
Contains:
 - BlockFiltering
 - BlockPurging
"""
from abc import ABC
from collections import defaultdict
from time import time
from typing import Tuple

import numpy as np
from tqdm.autonotebook import tqdm

from .block_building import AbstractBlockProcessing
from .datamodel import Block, Data
from .utils import create_entity_index, drop_single_entity_blocks, java_math_round

class AbstractBlockCleaning(AbstractBlockProcessing):

    def __init__(self) -> None:
        super().__init__()

    def report(self) -> None:
        """Prints Block Building method configuration
        """
        print(
            "Method name: " + self._method_name +
            "\nMethod info: " + self._method_info +
            "\nParameters: \n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in self._configuration().items()]) +
            "Runtime: {:2.4f} seconds".format(self.execution_time)
        )

class BlockFiltering(AbstractBlockCleaning):
    """Retains every entity in a subset of its smallest blocks.

        Filtering consists of 3 steps:
        - Blocks sort in ascending cardinality
        - Creation of Entity Index: inversed block dictionary
        - Retain every entity in ratio % of its smallest blocks
        - Blocks reconstruction
    """

    _method_name = "Block Filtering"
    _method_short_name: str = "BF"
    _method_info = "Retains every entity in a subset of its smallest blocks."

    def __init__(self, ratio: float = 0.8) -> None:
        super().__init__()
        if ratio > 1.0 or ratio < 0.0:
            raise AttributeError("Ratio is a number between 0.0 and 1.0")
        else:
            self.ratio = ratio
        self.entity_index: dict

    def __str__(self) -> str:
        print(self._method_name + self._method_info)
        print("Ratio: ", self.ratio)
        return super().__str__()

    def process(
            self,
            blocks: dict,
            data: Data,
            tqdm_disable: bool = False
    ) -> Tuple[dict, dict]:
        """Main method of Block Filtering.

        Args:
            blocks (dict): dict of keys to Blocks.
            data (Data): input dataset.
            tqdm_disable (bool, optional): disable progress bars. Defaults to False.

        Returns:
            Tuple[dict, dict]: dict of keys to Blocks, entity index (reversed blocks)
        """
        start_time, self.tqdm_disable, self.data = time(), tqdm_disable, data
        self._progress_bar = tqdm(
            total=3,
            desc=self._method_name,
            disable=self.tqdm_disable
        )
        sorted_blocks = _sort_blocks_cardinality(blocks, self.data.is_dirty_er)
        self._progress_bar.update(1)
        self.entity_index = create_entity_index(sorted_blocks, self.data.is_dirty_er)
        self._progress_bar.update(1)
        filtered_blocks = {}
        for entity_id, block_keys in self.entity_index.items():
            # Create new blocks from the entity index
            block_keys = list(block_keys)
            for key in list(block_keys[:java_math_round(self.ratio*float(len(block_keys)))]):
                filtered_blocks.setdefault(key, Block())
                # Entities ids start to 0 ... n-1 for 1st dataset
                # and n ... m for 2nd dataset
                _ = filtered_blocks[key].entities_D1.add(entity_id) if entity_id < self.data.dataset_limit \
                    else filtered_blocks[key].entities_D2.add(entity_id)
        self._progress_bar.update(1)
        new_blocks = drop_single_entity_blocks(filtered_blocks, self.data.is_dirty_er)
        self._progress_bar.close()
        self.num_of_blocks_dropped = len(blocks) - len(new_blocks)
        self.execution_time = time() - start_time
        self.blocks = new_blocks

        return self.blocks

    def _configuration(self) -> dict:
        return {
            "Ratio" : self.ratio
        }

class BlockPurging(AbstractBlockCleaning):
    """Discards the blocks exceeding a certain number of comparisons.
    """

    _method_name = "Block Purging"
    _method_short_name: str = "BP"
    _method_info = "Discards the blocks exceeding a certain number of comparisons."

    def __init__(self, smoothing_factor: float = 1.025) -> any:
        super().__init__()
        self.smoothing_factor: float = smoothing_factor
        self.max_comparisons_per_block: float

    def process(
            self,
            blocks: dict,
            data: Data,
            tqdm_disable: bool = False
    ) -> dict:
        """Main method of Block Purging.

        Args:
            blocks (dict): Blocks of entities.
            data (Data): Data module. Contains all the information about the dataset.
            tqdm_disable (bool, optional): Disable progress bar. Defaults to False.

        Returns:
            dict: Purged blocks.
        """
        self.tqdm_disable, self.data, start_time = tqdm_disable, data, time()
        if not blocks:
            raise AttributeError("Empty dict of blocks was given as input!")
        else:
            new_blocks = blocks.copy()
        self._progress_bar = tqdm(total=2*len(new_blocks), desc=self._method_name, disable=self.tqdm_disable)            
        self._set_threshold(new_blocks)
        new_blocks = dict(filter(self._cardinality_threshold, blocks.items()))
        self._progress_bar.close()
        self.execution_time = time() - start_time
        self.blocks = new_blocks

        return new_blocks

    def _cardinality_threshold(self, id_block_tuple) -> bool:
        self._progress_bar.update(1)
        return id_block_tuple[1].get_cardinality(self.data.is_dirty_er) <= self.max_comparisons_per_block

    def _set_threshold(self, blocks: dict) -> None:
        """Calculates the maximum number of comparisons per block, so in the next step to be purged.

        Args:
            blocks (dict): _description_
        """
        sorted_blocks = _sort_blocks_cardinality(blocks, self.data.is_dirty_er)
        distinct_comparisons_level = set(b.get_cardinality(self.data.is_dirty_er) \
                                        for _, b in sorted_blocks.items())
        block_assignments = np.empty([len(distinct_comparisons_level)])
        comparisons_level = np.empty([len(distinct_comparisons_level)])
        total_comparisons_per_level = np.empty([len(distinct_comparisons_level)])
        index = -1
        for _, block in sorted_blocks.items():
            if index == -1:
                index += 1
                comparisons_level[index] = block.get_cardinality(self.data.is_dirty_er)
                block_assignments[index] = 0
                total_comparisons_per_level[index] = 0
            elif block.get_cardinality(self.data.is_dirty_er) != comparisons_level[index]:
                index += 1
                comparisons_level[index] = block.get_cardinality(self.data.is_dirty_er)
                block_assignments[index] = block_assignments[index-1]
                total_comparisons_per_level[index] = total_comparisons_per_level[index-1]

            block_assignments[index] += block.get_size()
            total_comparisons_per_level[index] += block.get_cardinality(self.data.is_dirty_er)
            self._progress_bar.update(1)

        current_bc = current_cc = current_size = \
            previous_bc = previous_cc = previous_size = 0
        for i in range(len(block_assignments)-1, 0, -1):
            previous_size = current_size
            previous_bc = current_bc
            previous_cc = current_cc
            current_size = comparisons_level[i]
            current_bc = block_assignments[i]
            current_cc = total_comparisons_per_level[i]
            if current_bc * previous_cc < self.smoothing_factor * current_cc * previous_bc:
                break
        self.max_comparisons_per_block = previous_size

    def _satisfies_threshold(self, block: Block) -> bool:
        return block.get_cardinality(self.data.is_dirty_er) <= self.max_comparisons_per_block

    def _configuration(self) -> dict:
        return {
            "Smoothing factor" : self.smoothing_factor,
            "Max Comparisons per Block" : self.max_comparisons_per_block
        }

def _sort_blocks_cardinality(blocks: dict, is_dirty_er: bool) -> dict:
    return dict(sorted(blocks.items(), key=lambda x: x[1].get_cardinality(is_dirty_er)))



import itertools
import logging as log
import math
import re
import time
import pandas as pd
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Tuple

import nltk
import numpy as np
from tqdm.auto import tqdm

from .datamodel import Block, Data, PYJEDAIFeature
from .utils import (are_matching, drop_big_blocks_by_size,
                    drop_single_entity_blocks, get_blocks_cardinality)
from .evaluation import Evaluation

class AbstractBlockProcessing(PYJEDAIFeature):
    """Abstract class for the block building method
    """

    def __init__(self):
        super().__init__()
        self.blocks: dict
        self.attributes_1: list
        self.attributes_2: list
        self.num_of_blocks_dropped: int
        self.original_num_of_blocks: int
        self.sum_of_sizes: int = 0
        self.total_num_of_comparisons: int = 0
        self.min_block_size: int = None
        self.max_block_size: int = None
        self.min_block_comparisons: int = None
        self.max_block_comparisons: int = None
            
    def report(self) -> None:
        """Prints Block Building method configuration
        """
        print(
            "Method name: " + self._method_name +
            "\nMethod info: " + self._method_info +
            ("\nParameters: \n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in self._configuration().items()]) if self._configuration().items() else "\nParameters: Parameter-Free method\n") +
            "Attributes from D1:\n\t" + ', '.join(c for c in (self.attributes_1 if self.attributes_1 is not None \
                else self.data.dataset_1.columns)) +
            ("\nAttributes from D2:\n\t" + ', '.join(c for c in (self.attributes_2 if self.attributes_2 is not None \
                else self.data.dataset_2.columns)) if not self.data.is_dirty_er else "") +
            "\nRuntime: {:2.4f} seconds".format(self.execution_time)
        )

    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True,
                 with_stats: bool = False) -> any:

        if prediction is None:
            if self.blocks is None:
                raise AttributeError("Can not proceed to evaluation without build_blocks.")
            else:
                eval_blocks = self.blocks
        else:
            eval_blocks = prediction
            
        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " + 
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        entity_index = eval_obj._create_entity_index_from_blocks(eval_blocks)
        for _, (id1, id2) in self.data.ground_truth.iterrows():
            id1 = self.data._ids_mapping_1[id1]
            id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er else self.data._ids_mapping_2[id2]
            if id1 in entity_index and    \
                id2 in entity_index and are_matching(entity_index, id1, id2):
                true_positives += 1

        total_matching_pairs = get_blocks_cardinality(eval_blocks, self.data.is_dirty_er)
        eval_obj.calculate_scores(true_positives=true_positives, total_matching_pairs=total_matching_pairs)
        eval_result = eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)
        if with_stats:
            self.stats(eval_blocks)
        return eval_result

    def stats(self, blocks: dict) -> None:
        self.list_of_sizes = []
        self.entities_in_blocks = set()
        for block in blocks.values():
            self.sum_of_sizes += block.get_size()
            self.min_block_size = min(self.min_block_size, block.get_size()) if self.min_block_size else block.get_size()
            self.max_block_size = max(self.max_block_size, block.get_size()) if self.max_block_size else block.get_size()
            self.min_block_comparisons = min(self.min_block_comparisons, block.get_cardinality(self.data.is_dirty_er)) if self.min_block_comparisons else block.get_cardinality(self.data.is_dirty_er)
            self.max_block_comparisons = max(self.max_block_comparisons, block.get_cardinality(self.data.is_dirty_er)) if self.max_block_comparisons else block.get_cardinality(self.data.is_dirty_er)
            self.list_of_sizes.append(block.get_size())
            self.entities_in_blocks = self.entities_in_blocks.union(block.entities_D1)
            if not self.data.is_dirty_er:
                self.entities_in_blocks = self.entities_in_blocks.union(block.entities_D2)
            self.total_num_of_comparisons += block.get_cardinality(self.data.is_dirty_er)
        
        self.num_of_blocks = len(blocks)
        self.average_block_size = int(self.sum_of_sizes / self.num_of_blocks)
        self.list_of_sizes = sorted(self.list_of_sizes)
        median = self.list_of_sizes[int(len(self.list_of_sizes)/2)]
        print(
            "Statistics:" +
            "\n\tNumber of blocks: " + str(self.num_of_blocks) +
            "\n\tAverage block size: " + str(self.average_block_size) +
            "\n\tMedian block size: " + str(median) +
            "\n\tMax block size: " + str(self.max_block_size) +
            "\n\tMin block size: " + str(self.min_block_size) +
            "\n\tNumber of blocks dropped: " + str(self.num_of_blocks_dropped) +
            "\n\tNumber of comparisons: " + str(self.total_num_of_comparisons) +
            "\n\tMax comparisons per block: " + str(self.max_block_comparisons) +
            "\n\tMin comparisons per block: " + str(self.min_block_comparisons) +
            "\n\tEntities in blocks: " + str(len(self.entities_in_blocks))
        )
        print(u'\u2500' * 123)

    def export_to_df(
        self,
        blocks: dict
    ) -> pd.DataFrame:
        """creates a dataframe for the evaluation report

        Args:
            blocks (any): Predicted blocks
            data (Data): initial dataset

        Returns:
            pd.DataFrame: Dataframe predicted pairs (can be exported to csv)
        """
        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. \
                Data object mush have initialized with the ground-truth file")
        pairs_df = pd.DataFrame(columns=['id1', 'id2'])
        for _, block in blocks.items():
            if self.data.is_dirty_er:
                lblock = list(block.entities_D1)
                for i1 in range(0, len(lblock)):
                    for i2 in range(i1+1, len(lblock)):
                        id1 = self.data._gt_to_ids_reversed_1[lblock[i1]]
                        id2 = self.data._gt_to_ids_reversed_1[lblock[i2]] if self.data.is_dirty_er \
                            else self.data._gt_to_ids_reversed_2[lblock[i2]]
                        pairs_df = pd.concat([pairs_df, pd.DataFrame([{'id1':id1, 'id2':id2}], index=[0])], ignore_index=True)
            else:
                for i1 in block.entities_D1:
                    for i2 in block.entities_D2:
                        id1 = self.data._gt_to_ids_reversed_1[i1]
                        id2 = self.data._gt_to_ids_reversed_1[i2] if self.data.is_dirty_er \
                            else self.data._gt_to_ids_reversed_2[i2]
                        pairs_df = pd.concat([pairs_df, pd.DataFrame([{'id1':id1, 'id2':id2}], index=[0])], ignore_index=True)
        return pairs_df

class AbstractBlockBuilding(AbstractBlockProcessing):
    """Abstract class for the block building method
    """

    _method_name: str
    _method_info: str
    _method_short_name: str

    def __init__(self):
        super().__init__()
        self.blocks: dict
        self._progress_bar: tqdm
        self.attributes_1: list
        self.attributes_2: list
        self.execution_time: float
        self.data: Data
        self.list_of_sizes: list = []

    def build_blocks(
            self,
            data: Data,
            attributes_1: list = None,
            attributes_2: list = None,
            tqdm_disable: bool = False
    ) -> Tuple[dict, dict]:
        """Main method of Blocking in a dataset

            Args:
                data (Data): Data module that contaiins the processed dataset
                attributes_1 (list, optional): Attribute columns of the dataset 1 \
                    that will be processed. Defaults to None. \
                    If not provided, all attributes are slected.
                attributes_2 (list, optional): Attribute columns of the dataset 2. \
                    Defaults to None. If not provided, all attributes are slected.
                tqdm_disable (bool, optional): Disables all tqdm at processing. Defaults to False.

            Returns:
                Tuple[dict, dict]: Dictionary of blocks, Dict of entity index (reversed blocks).
        """

        _start_time = time.time()
        self.data, self.attributes_1, self.attributes_2 = data, attributes_1, attributes_2
        self._progress_bar = tqdm(
            total=data.num_of_entities, desc=self._method_name, disable=tqdm_disable
        )

        # TODO Text process function can be applied in this step (.apply)
        self._entities_d1 = data.dataset_1[attributes_1 if attributes_1 else data.attributes_1] \
                            .apply(" ".join, axis=1) \
                            .apply(self._tokenize_entity) \
                            .values.tolist()
                        # if attributes_1 else data.entities_d1.apply(self._tokenize_entity)

        self._all_tokens = set(itertools.chain.from_iterable(self._entities_d1))

        if not data.is_dirty_er:
            self._entities_d2 = data.dataset_2[attributes_2 if attributes_2 else data.attributes_2] \
                    .apply(" ".join, axis=1) \
                    .apply(self._tokenize_entity) \
                    .values.tolist()
            self._all_tokens.union(set(itertools.chain.from_iterable(self._entities_d2)))

        entity_id = itertools.count()
        blocks = {}
        
        for entity in self._entities_d1:
            eid = next(entity_id)
            for token in entity:
                blocks.setdefault(token, Block())
                blocks[token].entities_D1.add(eid)
            self._progress_bar.update(1)

        if not data.is_dirty_er:
            for entity in self._entities_d2:
                eid = next(entity_id)
                for token in entity:
                    blocks.setdefault(token, Block())
                    blocks[token].entities_D2.add(eid)
                self._progress_bar.update(1)

        self.original_num_of_blocks = len(blocks)
        self.blocks = self._clean_blocks(blocks)
        self.num_of_blocks_dropped = len(blocks) - len(self.blocks)
        self.execution_time = time.time() - _start_time
        self._progress_bar.close()

        return self.blocks

    def report(self) -> None:
        """Prints Block Building method configuration
        """
        print(
            "Method name: " + self._method_name +
            "\nMethod info: " + self._method_info +
            ("\nParameters: \n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in self._configuration().items()]) if self._configuration().items() else "\nParameters: Parameter-Free method\n") +
            "Attributes from D1:\n\t" + ', '.join(c for c in (self.attributes_1 if self.attributes_1 is not None \
                else self.data.dataset_1.columns)) +
            ("\nAttributes from D2:\n\t" + ', '.join(c for c in (self.attributes_2 if self.attributes_2 is not None \
                else self.data.dataset_2.columns)) if not self.data.is_dirty_er else "") +
            "\nRuntime: {:2.4f} seconds".format(self.execution_time)
        )

    @abstractmethod
    def _clean_blocks(self, blocks: dict) -> dict:
        pass

    @abstractmethod
    def _configuration(self) -> dict:
        pass
    
class StandardBlocking(AbstractBlockBuilding):
    """ Creates one block for every token in \
        the attribute values of at least two entities.
    """

    _method_name = "Standard Blocking"
    _method_short_name: str = "SB"
    _method_info = "Creates one block for every token in " + \
        "the attribute values of at least two entities."

    def __init__(self) -> any:
        super().__init__()

    def _tokenize_entity(self, entity: str) -> list:
        """Produces a list of workds of a given string

        Args:
            entity (str): String representation  of an entity

        Returns:
            list: List of words
        """
        return list(set(filter(None, re.split('[\\W_]', entity.lower()))))

    def _clean_blocks(self, blocks: dict) -> dict:
        """No cleaning"""
        return drop_single_entity_blocks(blocks, self.data.is_dirty_er)

    def _configuration(self) -> dict:
        """No configuration"""
        return {}

class QGramsBlocking(StandardBlocking):
    """ Creates one block for every q-gram that is extracted \
        from any token in the attribute values of any entity. \
            The q-gram must be shared by at least two entities.
    """

    _method_name = "Q-Grams Blocking"
    _method_short_name: str = "QGB"
    _method_info = "Creates one block for every q-gram that is extracted " + \
                    "from any token in the attribute values of any entity. " + \
                    "The q-gram must be shared by at least two entities."

    def __init__(
            self, qgrams: int = 6
    ) -> any:
        super().__init__()
        self.qgrams = qgrams

    def _tokenize_entity(self, entity) -> set:
        keys = set()
        for token in super()._tokenize_entity(entity):
            if len(token) < self.qgrams:
                keys.add(token)
            else:
                keys.update(''.join(qg) for qg in nltk.ngrams(token, n=self.qgrams))
        return keys

    def _clean_blocks(self, blocks: dict) -> dict:
        return drop_single_entity_blocks(blocks, self.data.is_dirty_er)

    def _configuration(self) -> dict:
        return {
            "Q-Gramms" : self.qgrams
        }

class SuffixArraysBlocking(StandardBlocking):
    """ It creates one block for every suffix that appears \
        in the attribute value tokens of at least two entities.
    """

    _method_name = "Suffix Arrays Blocking"
    _method_short_name: str = "SAB"
    _method_info = "Creates one block for every suffix that appears in the " + \
        "attribute value tokens of at least two entities."

    def __init__(
            self,
            suffix_length: int = 6,
            max_block_size: int = 53
    ) -> any:
        super().__init__()
        self.suffix_length, self.max_block_size = suffix_length, max_block_size

    def _tokenize_entity(self, entity) -> set:
        keys = set()
        for token in super()._tokenize_entity(entity):
            if len(token) < self.suffix_length:
                keys.add(token)
            else:
                for length in range(0, len(token) - self.suffix_length + 1):
                    keys.add(token[length:])
        return keys

    def _clean_blocks(self, blocks: dict) -> dict:
        return drop_big_blocks_by_size(blocks, self.max_block_size, self.data.is_dirty_er)

    def _configuration(self) -> dict:
        return {
            "Suffix length" : self.suffix_length,
            "Maximum Block Size" : self.max_block_size
        }

class ExtendedSuffixArraysBlocking(StandardBlocking):
    """ It creates one block for every substring \
        (not just suffix) that appears in the tokens of at least two entities.
    """

    _method_name = "Extended Suffix Arrays Blocking"
    _method_short_name: str = "ESAB"
    _method_info = "Creates one block for every substring (not just suffix) " + \
        "that appears in the tokens of at least two entities."

    def __init__(
            self,
            suffix_length: int = 6,
            max_block_size: int = 39
    ) -> any:
        super().__init__()
        self.suffix_length, self.max_block_size = suffix_length, max_block_size

    def _tokenize_entity(self, entity) -> set:
        keys = set()
        for token in super()._tokenize_entity(entity):
            keys.add(token)
            if len(token) > self.suffix_length:
                for current_size in range(self.suffix_length, len(token)): 
                    for letters in list(nltk.ngrams(token, n=current_size)):
                        keys.add("".join(letters))
        return keys

    def _clean_blocks(self, blocks: dict) -> dict:
        return drop_big_blocks_by_size(blocks, self.max_block_size, self.data.is_dirty_er)

    def _configuration(self) -> dict:
        return {
            "Suffix length" : self.suffix_length,
            "Maximum Block Size" : self.max_block_size
        }

class ExtendedQGramsBlocking(StandardBlocking):
    """It creates one block for every combination of q-grams that represents at least two entities.
    The q-grams are extracted from any token in the attribute values of any entity.
    """

    _method_name = "Extended QGramsBlocking"
    _method_short_name: str = "EQGB"
    _method_info = "Creates one block for every substring (not just suffix) " + \
        "that appears in the tokens of at least two entities."

    def __init__(
            self,
            qgrams: int = 6,
            threshold: float = 0.95
    ) -> any:
        super().__init__()
        self.threshold: float = threshold
        self.MAX_QGRAMS: int = 15
        self.qgrams = qgrams

    def _tokenize_entity(self, entity) -> set:
        keys = set()
        for token in super()._tokenize_entity(entity):
            if len(token) < self.qgrams:
                keys.add(token)
            else:   
                qgrams = [''.join(qgram) for qgram in nltk.ngrams(token, n=self.qgrams)]
                if len(qgrams) == 1:
                    keys.update(qgrams)
                else:
                    if len(qgrams) > self.MAX_QGRAMS:
                        qgrams = qgrams[:self.MAX_QGRAMS]

                    minimum_length = max(1, math.floor(len(qgrams) * self.threshold))
                    for i in range(minimum_length, len(qgrams) + 1):
                        keys.update(self._qgrams_combinations(qgrams, i))

        return keys

    def _qgrams_combinations(self, sublists: list, sublist_length: int) -> list:
        if sublist_length == 0 or len(sublists) < sublist_length:
            return []

        remaining_elements = sublists.copy()
        last_sublist = remaining_elements.pop(len(sublists)-1)

        combinations_exclusive_x = self._qgrams_combinations(remaining_elements, sublist_length)
        combinations_inclusive_x = self._qgrams_combinations(remaining_elements, sublist_length-1)

        resulting_combinations = combinations_exclusive_x.copy() if combinations_exclusive_x else []

        if not combinations_inclusive_x: # is empty
            resulting_combinations.append(last_sublist)
        else:
            for combination in combinations_inclusive_x:
                resulting_combinations.append(combination+last_sublist)

        return resulting_combinations

    def _clean_blocks(self, blocks: dict) -> dict:
        return drop_single_entity_blocks(blocks, self.data.is_dirty_er)

    def _configuration(self) -> dict:
        return {
            "Q-Gramms" : self.qgrams,
            "Threshold" : self.threshold
        }



import os
import sys

import pandas as pd
sys.path.append(os.path.join(os.path.dirname(__file__), '../'))

# --------------------------------- #
# Datamodel
# --------------------------------- #

from pyjedai.datamodel import Data

dirty_data = Data(
    dataset_1=pd.read_csv("data/der/cora/cora.csv", sep='|'),
    id_column_name_1='Entity Id',
    ground_truth=pd.read_csv("data/der/cora/cora_gt.csv", sep='|', header=None),
)
dirty_data.print_specs()

def test_datamodel_dirty_er():
    assert dirty_data is not None

clean_clean_data = Data(
    dataset_1=pd.read_csv("data/test/ccer/abt_100.csv", sep='|', engine='python').astype(str),
    id_column_name_1='id',
    dataset_2=pd.read_csv("data/test/ccer/buy_100.csv", sep='|', engine='python').astype(str),
    id_column_name_2='id',
    ground_truth=pd.read_csv("data/test/ccer/gt_100.csv", sep='|', engine='python')
)
clean_clean_data.print_specs()

def test_datamodel_clean_clean_er():
    assert clean_clean_data is not None

# --------------------------------- #
# Block building
# --------------------------------- #

from pyjedai.block_building import StandardBlocking
dblocks = StandardBlocking().build_blocks(dirty_data)
ccblocks = StandardBlocking().build_blocks(clean_clean_data)

# --------------------------------- #
# Block cleaning
# --------------------------------- #
from pyjedai.block_cleaning import BlockFiltering
dblocks = BlockFiltering().process(dblocks, dirty_data)
ccblocks =BlockFiltering().process(ccblocks, clean_clean_data)

from pyjedai.block_cleaning import BlockPurging
dblocks = BlockPurging().process(dblocks, dirty_data)
ccblocks =BlockPurging().process(ccblocks, clean_clean_data)

# --------------------------------- #
# Comparison cleaning
# --------------------------------- #

def test_WeightedEdgePruning():
    from pyjedai.comparison_cleaning import WeightedEdgePruning
    assert WeightedEdgePruning().process(dblocks, dirty_data) is not None
    assert WeightedEdgePruning().process(ccblocks, clean_clean_data) is not None

def test_WeightedNodePruning():
    from pyjedai.comparison_cleaning import WeightedNodePruning
    assert WeightedNodePruning().process(dblocks, dirty_data) is not None
    assert WeightedNodePruning().process(ccblocks, clean_clean_data) is not None

def test_CardinalityEdgePruning():
    from pyjedai.comparison_cleaning import CardinalityEdgePruning
    assert CardinalityEdgePruning().process(dblocks, dirty_data) is not None
    assert CardinalityEdgePruning().process(ccblocks, clean_clean_data) is not None

def test_CardinalityNodePruning():
    from pyjedai.comparison_cleaning import CardinalityNodePruning
    assert CardinalityNodePruning().process(dblocks, dirty_data) is not None
    assert CardinalityNodePruning().process(ccblocks, clean_clean_data) is not None

def test_BLAST():
    from pyjedai.comparison_cleaning import BLAST
    assert BLAST().process(dblocks, dirty_data) is not None
    assert BLAST().process(ccblocks, clean_clean_data) is not None

def test_ReciprocalCardinalityNodePruning():
    from pyjedai.comparison_cleaning import ReciprocalCardinalityNodePruning
    assert ReciprocalCardinalityNodePruning().process(dblocks, dirty_data) is not None
    assert ReciprocalCardinalityNodePruning().process(ccblocks, clean_clean_data) is not None

def test_ReciprocalWeightedNodePruning():
    from pyjedai.comparison_cleaning import ReciprocalWeightedNodePruning
    assert ReciprocalWeightedNodePruning().process(dblocks, dirty_data) is not None
    assert ReciprocalWeightedNodePruning().process(ccblocks, clean_clean_data) is not None

def test_ComparisonPropagation():
    from pyjedai.comparison_cleaning import ComparisonPropagation
    assert ComparisonPropagation().process(dblocks, dirty_data) is not None
    assert ComparisonPropagation().process(ccblocks, clean_clean_data) is not None



"""Entity Matching Prioritization Module
"""
import numpy as np
from time import time
import matplotlib.pyplot as plt
from .matching import EntityMatching
from .comparison_cleaning import (
    ComparisonPropagation,
    ProgressiveCardinalityEdgePruning,
    ProgressiveCardinalityNodePruning,
    GlobalProgressiveSortedNeighborhood,
    LocalProgressiveSortedNeighborhood,
    ProgressiveEntityScheduling)
from .joins import TopKJoin
from .vector_based_blocking import EmbeddingsNNBlockBuilding

from networkx import Graph
from py_stringmatching.similarity_measure.cosine import Cosine
from py_stringmatching.similarity_measure.dice import Dice
from py_stringmatching.similarity_measure.generalized_jaccard import \
    GeneralizedJaccard
from py_stringmatching.similarity_measure.jaccard import Jaccard
from py_stringmatching.similarity_measure.jaro import Jaro
from py_stringmatching.similarity_measure.levenshtein import Levenshtein
from py_stringmatching.similarity_measure.overlap_coefficient import \
    OverlapCoefficient
from py_stringmatching.tokenizer.qgram_tokenizer import QgramTokenizer
from py_stringmatching.tokenizer.whitespace_tokenizer import \
    WhitespaceTokenizer
from sklearn.metrics.pairwise import pairwise_distances
from tqdm.autonotebook import tqdm

from .datamodel import Data, PYJEDAIFeature
from .evaluation import Evaluation
from .matching import EntityMatching
from .comparison_cleaning import AbstractMetablocking
from queue import PriorityQueue
from random import sample
from abc import abstractmethod
from typing import Tuple, List
from .utils import (
    SubsetIndexer,
    DatasetScheduler,
    EntityScheduler,
    is_infinite,
    PredictionData,
    reverse_data_indexing,
    reverse_blocks_entity_indexing,
    sorted_enumerate,
    canonical_swap,
    WordQgramTokenizer,
    cosine,
    get_qgram_from_tokenizer_name,
    FrequencyEvaluator)
import pandas as pd
import os
from whoosh.fields import TEXT, Schema, ID
from whoosh.index import create_in
from whoosh import qparser
from whoosh.scoring import TF_IDF, Frequency, PL2, BM25F
from collections import defaultdict
import sys
from faiss import METRIC_INNER_PRODUCT, METRIC_L2


# Directory where the whoosh index is stored
INDEXER_DIR='.indexer'

metrics_mapping = {
    'edit_distance': Levenshtein(),
    'cosine' : Cosine(),
    'jaro' : Jaro(),
    'jaccard' : Jaccard(),
    'generalized_jaccard' : GeneralizedJaccard(),
    'dice': Dice(),
    'overlap_coefficient' : OverlapCoefficient(),
}

vector_metrics_mapping = {
    'cosine': cosine
}

string_metrics = [
    'jaro', 'edit_distance'
]

set_metrics = [
    'cosine', 'dice', 'generalized_jaccard', 'jaccard', 'overlap_coefficient'
]

vector_metrics = [ 
    'cosine', 'dice', 'jaccard'
]

whoosh_index_metrics = [
    'TF-IDF', 'Frequency', 'PL2', 'BM25F'
] 

faiss_metrics = [
    'cosine', 'euclidean'
]

magellan_metrics = string_metrics + set_metrics
available_metrics = magellan_metrics + vector_metrics + whoosh_index_metrics + faiss_metrics

#
# Tokenizers
#
char_qgram_tokenizers = { 'char_'+ str(i) + 'gram':i for i in range(1, 7) }
word_qgram_tokenizers = { 'word_'+ str(i) + 'gram':i for i in range(1, 7) }
magellan_tokenizers = ['white_space_tokenizer']

tfidf_tokenizers = [ 'tfidf_' + cq for cq in char_qgram_tokenizers.keys() ] + \
                    [ 'tfidf_' + wq for wq in word_qgram_tokenizers.keys() ]

tf_tokenizers = [ 'tf_' + cq for cq in char_qgram_tokenizers.keys() ] + \
                    [ 'tf_' + wq for wq in word_qgram_tokenizers.keys() ]
                        
boolean_tokenizers = [ 'boolean_' + cq for cq in char_qgram_tokenizers.keys() ] + \
                        [ 'boolean_' + wq for wq in word_qgram_tokenizers.keys() ]

vector_tokenizers = tfidf_tokenizers + tf_tokenizers + boolean_tokenizers

available_tokenizers = [key for key in char_qgram_tokenizers] + [key for key in word_qgram_tokenizers] + magellan_tokenizers + vector_tokenizers

class ProgressiveMatching(EntityMatching):
    """Applies the matching process to a subset of available pairs progressively 
    """

    _method_name: str = "Progressive Matching"
    _method_info: str = "Applies the matching process to a subset of available pairs progressively "
    def __init__(
            self,
            similarity_function: str = 'dice',
            tokenizer: str = 'white_space_tokenizer',
            vectorizer : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.0,
            tokenizer_return_unique_values = True, # unique values or not
            attributes: any = None,
        ) -> None:

        super().__init__(metric=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)
        self.similarity_function : str = similarity_function
        
    def predict(self,
            data: Data,
            blocks: dict,
            budget: int = 0,
            algorithm : str = 'HB',
            indexing : str = 'inorder',
            comparison_cleaner: AbstractMetablocking = None,
            tqdm_disable: bool = False,
            emit_all_tps_stop : bool = False) -> List[Tuple[float, int, int]]:
        """Main method of  progressive entity matching. Inputs a set of blocks and outputs a list \
           that contains duplets of ids corresponding to candidate pairs to emit.
            Args:
                blocks (dict): blocks of entities
                data (Data): dataset module
                tqdm_disable (bool, optional): Disables progress bar. Defaults to False.
                method (str) : DFS/BFS/Hybrid approach for specified algorithm
                emit_all_tps_stop (bool) : Stop emission once all true positives are found
            Returns:
                networkx.Graph: entity ids (nodes) and similarity scores between them (edges)
        """
        start_time = time()
        self.tqdm_disable = tqdm_disable
        self._budget : int = budget
        self._indexing : str = indexing
        self._comparison_cleaner: AbstractMetablocking = comparison_cleaner
        self._algorithm : str= algorithm
        self._emit_all_tps_stop : bool = emit_all_tps_stop
        self.duplicate_emitted : dict = None if not self._emit_all_tps_stop else {}
        self._prediction_data : PredictionData = None
        self.data : Data = data
        self.duplicate_of = data.duplicate_of
        self.scheduler : DatasetScheduler = None

        if not blocks:
            raise ValueError("Empty blocks structure")
        
        if self.data.is_dirty_er and self._indexing == 'bilateral':
            raise ValueError("Cannot apply bilateral indexing to dirty Entity Resolution (single dataset)")
            
        _inorder_blocks = blocks  
        self._pairs_top_score : dict = defaultdict(lambda: -1)
        all_blocks = list(blocks.values())
        self._progress_bar = tqdm(total=len(blocks),
                        desc=self._method_name,
                        disable=self.tqdm_disable)
        
        if(indexing == 'bilateral'): self._indexing = 'inorder'
        if(self._indexing == 'inorder'):
            if 'Block' in str(type(all_blocks[0])):
                self._predict_raw_blocks(blocks)
            elif isinstance(all_blocks[0], set):
                if(self._comparison_cleaner == None):
                    raise AttributeError("No precalculated weights were given from the CC step") 
                self._predict_prunned_blocks(blocks)
            else:
                raise AttributeError("Wrong type of Blocks")
            self._schedule_candidates()
            
            
        if(indexing == 'bilateral'): self._indexing = 'reverse'
        if(self._indexing == 'reverse'):
            _reverse_blocks = reverse_blocks_entity_indexing(_inorder_blocks, self.data)
            self.data = reverse_data_indexing(self.data)
            if 'Block' in str(type(all_blocks[0])):
                self._predict_raw_blocks(_reverse_blocks)
            elif isinstance(all_blocks[0], set):
                if(self._comparison_cleaner == None):
                    raise AttributeError("No precalculated weights were given from the CC step") 
                self._predict_prunned_blocks(_reverse_blocks)
            else:
                raise AttributeError("Wrong type of Blocks")
            self._schedule_candidates()
        
        self._gather_top_pairs()
        self.execution_time = time() - start_time
        self._progress_bar.close()
        
        return self.pairs
    
    
    def _store_id_mappings(self) -> None:
        """Stores the mapping [Workflow ID -> Dataframe ID] for the current indexing phase
        """
        if(self._indexing == "inorder"):
            self._inorder_d1_id = self.data._gt_to_ids_reversed_1
            self._inorder_d2_id = self.data._gt_to_ids_reversed_2  
        if(self._indexing == "reverse"):
            self._reverse_d1_id = self.data._gt_to_ids_reversed_1
            self._reverse_d2_id = self.data._gt_to_ids_reversed_2 
      
    def _schedule_candidates(self) -> None:
        """Translates the workflow identifiers back into dataframe identifiers
           Populates the dataset scheduler with the candidate pairs of the current indexing stage
        """
        self.scheduler = DatasetScheduler(budget=float('inf') if self._emit_all_tps_stop else self._budget, global_top=(self._algorithm=="TOP")) if self.scheduler == None else self.scheduler
        self._store_id_mappings()
        
        for score, entity, candidate in self.pairs:            
            # entities of first and second dataframe in the context of the current indexing
            d1_entity, d2_entity = (entity, candidate) if(entity < candidate) else (candidate, entity)
            d1_map, d2_map = (self._inorder_d1_id, self._inorder_d2_id) if (self._indexing == 'inorder') else (self._reverse_d1_id, self._reverse_d2_id)
            
            # print(f"#############################################################")
            # print(f"Score: {score}")
            # print(f"---------------Workflow IDs [{self._indexing}]---------------")
            # print(f"Entity: {entity}")
            # print(f"Candidate: {candidate}")
            # print(f"---------------Workflow IDs [D1 context Ent First]---------------")
            # print(f"D1 Entity: {d1_entity}")
            # print(f"D2 Entity: {d2_entity}")
            
            # the dataframe ids of the entities from first and second dataset in the context of indexing
            d1_entity_df_id, d2_entity_df_id = (d1_map[d1_entity], d2_map[d2_entity])
            _inorder_d1_entity_df_id, _inorder_d2_entity_df_id = (d1_entity_df_id, d2_entity_df_id) if (self._indexing == 'inorder') else (d2_entity_df_id, d1_entity_df_id)
            if(self._emit_all_tps_stop and _inorder_d2_entity_df_id in self.duplicate_of[_inorder_d1_entity_df_id]):
                self.duplicate_emitted[(_inorder_d1_entity_df_id, _inorder_d2_entity_df_id)] = False          
            
            # in the case of reverse indexing stage, adjust the workflow identifiers of the entities so we can differ them from inorder entity ids
            d1_entity = d1_entity if(self._indexing == 'inorder') else d1_entity + self.data.num_of_entities
            d2_entity = d2_entity if(self._indexing == 'inorder') else d2_entity + self.data.num_of_entities
            
            # print(f"---------------Dataframe IDs [{self._indexing}]---------------")
            # print(f"D1 Entity DF ID: {d1_entity_df_id}")
            # print(f"D2 Entity DF ID: {d2_entity_df_id}")
            # print(f"---------------Inorder Dataframe IDs [{self._indexing}]---------------")
            # print(f"Inorder D1 Entity DF ID: {_inorder_d1_entity_df_id}")
            # print(f"Inorder D2 Entity DF ID: {_inorder_d2_entity_df_id}")
            # print(f"---------------Scheduler IDs [D1 context Ent First]---------------")
            # print(f"D1 Entity: {d1_entity}")
            # print(f"D2 Entity: {d2_entity}")
            # if(_inorder_d2_entity_df_id in self.duplicate_of[_inorder_d1_entity_df_id]):
            #     print("^ THIS IS A TRUE POSITIVE ^") 
            # we want entities to be inserted in D1 -> D2 order (current context e.x. reverse) which translates to D2 -> D1 order (reverse context e.x. inorder)
            self.scheduler._insert_entity_neighbor(d1_entity, d2_entity, score)
            
    def _inorder_phase_entity(self, id : int) -> bool:
        """Given identifier corresponds to an entity proposed in the inorder indexing phase

        Args:
            id (int): Identifier

        Returns:
            bool: Identifier proposed in the inorder phase
        """
        return id < self.data.num_of_entities
    
    def _retrieve_entity_df_id(self, id : int) -> int:
        """Returns the corresponding id in the dataframe of the given entity id in the context of its indexing phase 

        Args:
            id (int): Workflow Identifier

        Returns:
            int: Dataframe Identifier
        """
        _workflow_id : int
        _df_id_of : dict
        if(self._inorder_phase_entity(id)):
            _workflow_id = id
            _df_id_of = self._inorder_d1_id if (_workflow_id < len(self._inorder_d1_id)) else self._inorder_d2_id
        else:
            _workflow_id = id - self.data.num_of_entities
            _df_id_of = self._reverse_d1_id if (_workflow_id < len(self._reverse_d1_id)) else self._reverse_d2_id
            
        return _df_id_of[_workflow_id]    
    
    def _gather_top_pairs(self) -> None:
        """Emits the pairs from the scheduler based on the defined algorithm
        """
        self.scheduler._sort_neighborhoods_by_avg_weight()
        self.pairs = self.scheduler._emit_pairs(method=self._algorithm, data=self.data)
        
        _identified_pairs = []
        for score, entity, candidate in self.pairs:
            _inorder_entities : bool = self._inorder_phase_entity(entity)
            entity, candidate = (self._retrieve_entity_df_id(entity), self._retrieve_entity_df_id(candidate))
            entity, candidate = (entity, candidate) if _inorder_entities else (candidate, entity)
            _identified_pairs.append((score, entity, candidate))
            
        self.pairs = _identified_pairs

    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:

        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " +
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        total_matching_pairs = prediction.number_of_edges()
        for _, (id1, id2) in self.data.ground_truth.iterrows():
            id1 = self.data._ids_mapping_1[id1]
            id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er \
                                                else self.data._ids_mapping_2[id2]
            if (id1 in prediction and id2 in prediction[id1]) or   \
                 (id2 in prediction and id1 in prediction[id2]):
                true_positives += 1

        eval_obj.calculate_scores(true_positives=true_positives,
                                  total_matching_pairs=total_matching_pairs)
        return eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)
    
    def get_prediction_data(self) -> PredictionData:
        if(self._prediction_data is None):
            raise ValueError("Pairs not emitted yet - No Data to show")
        return self._prediction_data
    
    def get_total_emissions(self) -> int:
        return self.get_prediction_data().get_total_emissions()
    
    def get_cumulative_recall(self) -> float:
        return self.get_prediction_data().get_cumulative_recall()
    
    def get_normalized_auc(self) -> float:
        return self.get_prediction_data().get_normalized_auc()
    
    def set_prediction_data(self, prediction_data : PredictionData):
        self._prediction_data : PredictionData = prediction_data
class BlockIndependentPM(ProgressiveMatching):
    """Applies the matching process to a subset of available pairs progressively 
    """

    _method_name: str = "Progressive Matching"
    _method_info: str = "Applies the matching process to a subset of available pairs progressively "

    def __init__(
            self,
            similarity_function: str = 'dice',
            tokenizer: str = 'white_space_tokenizer',
            vectorizer : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.0,
            tokenizer_return_unique_values = True, # unique values or not
            attributes: any = None,
        ) -> None:

        super().__init__(similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)
        
    def predict(self,
            data: Data,
            blocks: dict,
            budget: int = 0,
            algorithm : str = 'HB',
            indexing : str = 'inorder',
            comparison_cleaner: AbstractMetablocking = None,
            tqdm_disable: bool = False,
            emit_all_tps_stop : bool = False) -> List[Tuple[float, int, int]]:
        """Main method of  progressive entity matching. Inputs a set of blocks and outputs a list \
           that contains duplets of ids corresponding to candidate pairs to emit.
            Args:
                blocks (dict): blocks of entities
                data (Data): dataset module
                tqdm_disable (bool, optional): Disables progress bar. Defaults to False.
                method (str) : DFS/BFS/Hybrid approach for specified algorithm
                emit_all_tps_stop (bool) : Stop emission once all true positives are found
            Returns:
                networkx.Graph: entity ids (nodes) and similarity scores between them (edges)
        """
        start_time = time()
        self.tqdm_disable = tqdm_disable
        self._budget : int = budget
        self._indexing : str = indexing
        self._comparison_cleaner: AbstractMetablocking = comparison_cleaner
        self._algorithm : str= algorithm
        self._emit_all_tps_stop : bool = emit_all_tps_stop
        self.duplicate_emitted : dict = None 
        self._prediction_data : PredictionData = None
        self.data : Data = data
        self.duplicate_of = data.duplicate_of
        self.scheduler : DatasetScheduler = None
        
        if self.data.is_dirty_er and self._indexing == 'bilateral':
            raise ValueError("Cannot apply bilateral indexing to dirty Entity Resolution (single dataset)")
            
        _inorder_blocks = blocks  
        self._pairs_top_score : dict = defaultdict(lambda: -1)
        all_blocks = list(blocks.values()) if blocks is not None else None
        self._progress_bar = tqdm(total=len(blocks) if blocks is not None else 0,
                        desc=self._method_name,
                        disable=self.tqdm_disable)
        
        if(indexing == 'bilateral'): self._indexing = 'inorder'
        if(self._indexing == 'inorder'):
            if all_blocks is None or 'Block' in str(type(all_blocks[0])):
                self._predict_raw_blocks(blocks)
            elif isinstance(all_blocks[0], set):
                if(self._comparison_cleaner == None):
                    raise AttributeError("No precalculated weights were given from the CC step") 
                self._predict_prunned_blocks(blocks)
            else:
                raise AttributeError("Wrong type of Blocks")
            self._schedule_candidates()
            
            
        if(indexing == 'bilateral'): self._indexing = 'reverse'
        if(self._indexing == 'reverse'):
            _reverse_blocks = reverse_blocks_entity_indexing(_inorder_blocks, self.data)
            self.data = reverse_data_indexing(self.data)
            if all_blocks is None or 'Block' in str(type(all_blocks[0])):
                self._predict_raw_blocks(_reverse_blocks)
            elif isinstance(all_blocks[0], set):
                if(self._comparison_cleaner == None):
                    raise AttributeError("No precalculated weights were given from the CC step") 
                self._predict_prunned_blocks(_reverse_blocks)
            else:
                raise AttributeError("Wrong type of Blocks")
            self._schedule_candidates()
        
        self._gather_top_pairs()
        self.execution_time = time() - start_time
        self._progress_bar.close()
        
        return self.pairs

class HashBasedProgressiveMatching(ProgressiveMatching):
    """Applies hash based candidate graph prunning, sorts retained comparisons and applies Progressive Matching
    """

    _method_name: str = "Hash Based Progressive Matching"
    _method_info: str = "Applies hash based candidate graph prunning, sorts retained comparisons and applies Progressive Matching"

    def __init__(
        self,
        weighting_scheme: str = 'X2',
        similarity_function: str = 'dice',
        tokenizer: str = 'white_space_tokenizer',
        vectorizer : str = None,
        qgram : int = 1,
        similarity_threshold: float = 0.0,
        tokenizer_return_unique_values = True, # unique values or not
        attributes: any = None,
    ) -> None:

        super().__init__(similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)
        self._weighting_scheme : str = weighting_scheme

class GlobalTopPM(HashBasedProgressiveMatching):
    """Applies Progressive CEP, sorts retained comparisons and applies Progressive Matching
    """

    _method_name: str = "Global Top Progressive Matching"
    _method_info: str = "Applies Progressive CEP, sorts retained comparisons and applies Progressive Matching"

    def __init__(
        self,
        weighting_scheme: str = 'X2',
        similarity_function: str = 'dice',
        tokenizer: str = 'white_space_tokenizer',
        vectorizer : str = None,
        qgram : int = 1,
        similarity_threshold: float = 0.0,
        tokenizer_return_unique_values = True, # unique values or not
        attributes: any = None,
    ) -> None:

        super().__init__(weighting_scheme=weighting_scheme,
                        similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)

    def _predict_raw_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        self.pairs = Graph()
        pcep : ProgressiveCardinalityEdgePruning = ProgressiveCardinalityEdgePruning(self._weighting_scheme, self._budget)
        candidates : dict = pcep.process(blocks=blocks, data=self.data, tqdm_disable=True, cc=None, emit_all_tps_stop=self._emit_all_tps_stop)
        self.blocks = candidates

        for entity_id, candidate_ids in candidates.items():
            for candidate_id in candidate_ids:
                self._insert_to_graph(entity_id, candidate_id, pcep.get_precalculated_weight(entity_id, candidate_id))
         
        self.pairs.edges = sorted(self.pairs.edges(data=True), key=lambda x: x[2]['weight'], reverse=True)
        self.pairs = [(edge[2]['weight'], edge[0], edge[1]) for edge in self.pairs.edges]
        return self.pairs


    def _predict_prunned_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        self.pairs = Graph()
        pcep : ProgressiveCardinalityEdgePruning = ProgressiveCardinalityEdgePruning(self._weighting_scheme, self._budget)
        candidates : dict = pcep.process(blocks=blocks, data=self.data, tqdm_disable=True, cc=self._comparison_cleaner, emit_all_tps_stop=self._emit_all_tps_stop)
        self.blocks = candidates

        for entity_id, candidate_ids in candidates.items():
            for candidate_id in candidate_ids:
                self._insert_to_graph(entity_id, candidate_id, self._comparison_cleaner.get_precalculated_weight(entity_id, candidate_id))

        self.pairs.edges = sorted(self.pairs.edges(data=True), key=lambda x: x[2]['weight'], reverse=True)
        self.pairs = [(edge[2]['weight'], edge[0], edge[1]) for edge in self.pairs.edges]
        return self.pairs

class LocalTopPM(HashBasedProgressiveMatching):
    """Applies Progressive CNP, sorts retained comparisons and applies Progressive Matching
    """

    _method_name: str = "Global Top Progressive Matching"
    _method_info: str = "Applies Progressive CNP, sorts retained comparisons and applies Progressive Matching"

    def __init__(
        self,
        weighting_scheme: str = 'X2',
        similarity_function: str = 'dice',
        number_of_nearest_neighbors: int = 10,
        tokenizer: str = 'white_space_tokenizer',
        vectorizer : str = None,
        qgram : int = 1,
        similarity_threshold: float = 0.0,
        tokenizer_return_unique_values = True, # unique values or not
        attributes: any = None,
    ) -> None:

        super().__init__(weighting_scheme=weighting_scheme,
                        similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)
        self._number_of_nearest_neighbors : int = number_of_nearest_neighbors
        
    def _predict_raw_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        self.pairs = Graph()
        pcnp : ProgressiveCardinalityNodePruning = ProgressiveCardinalityNodePruning(weighting_scheme=self._weighting_scheme, budget=self._budget)
        candidates : dict = pcnp.process(blocks=blocks, data=self.data, number_of_nearest_neighbors=self._number_of_nearest_neighbors, tqdm_disable=True, cc=None, emit_all_tps_stop=self._emit_all_tps_stop)
        self.blocks = candidates
        
        for entity_id, candidate_ids in candidates.items():
            for candidate_id in candidate_ids:
                self._insert_to_graph(entity_id, candidate_id, pcnp.get_precalculated_weight(entity_id, candidate_id))

        self.pairs.edges = sorted(self.pairs.edges(data=True), key=lambda x: x[2]['weight'], reverse=True)
        self.pairs = [(edge[2]['weight'], edge[0], edge[1]) for edge in self.pairs.edges]
        return self.pairs

    def _predict_prunned_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        self.pairs = Graph()
        pcnp : ProgressiveCardinalityNodePruning = ProgressiveCardinalityNodePruning(self._weighting_scheme, self._budget)
        candidates : dict = pcnp.process(blocks=blocks, data=self.data, number_of_nearest_neighbors=self._number_of_nearest_neighbors, tqdm_disable=True, cc=self._comparison_cleaner, emit_all_tps_stop=self._emit_all_tps_stop)
        self.blocks = candidates

        for entity_id, candidate_ids in candidates.items():
            for candidate_id in candidate_ids:
                self._insert_to_graph(entity_id, candidate_id, self._comparison_cleaner.get_precalculated_weight(entity_id, candidate_id))

        self.pairs.edges = sorted(self.pairs.edges(data=True), key=lambda x: x[2]['weight'], reverse=True) 
        self.pairs = [(edge[2]['weight'], edge[0], edge[1]) for edge in self.pairs.edges]
        return self.pairs


class EmbeddingsNNBPM(BlockIndependentPM):
    """Utilizes/Creates entity embeddings, constructs neighborhoods via NN Approach and applies Progressive Matching
    """

    _method_name: str = "Embeddings NN Blocking Based Progressive Matching"
    _method_info: str = "Utilizes/Creates entity embeddings, constructs neighborhoods via NN Approach and applies Progressive Matching"

    def __init__(
            self,
            language_model: str = 'bert',
            number_of_nearest_neighbors: int = 10,
            similarity_search: str = 'faiss',
            vector_size: int = 300,
            num_of_clusters: int = 5,
            similarity_function: str = 'cosine',
            tokenizer: str = 'white_space_tokenizer',
            vectorizer : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.0,
            tokenizer_return_unique_values = True, # unique values or not
            attributes: any = None
        ) -> None:

        super().__init__(similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)
        
        self._language_model : str = language_model
        self._number_of_nearest_neighbors : int = number_of_nearest_neighbors
        self._similarity_search : str = similarity_search
        self._vector_size : int = vector_size
        self._num_of_clusters : int = num_of_clusters

    def _top_pair_emission(self) -> None:
        """Applies global sorting to all entity pairs produced by NN,
           and returns pairs based on distance in ascending order
        """
        self.pairs = []
        n, k = self.neighbors.shape

        for i in range(n):
            entity_id = self.ennbb._si.d1_retained_ids[i] if self.data.is_dirty_er else self.ennbb._si.d2_retained_ids[i]
            for j in range(k):
                if self.neighbors[i][j] != -1:
                    candidate_id = self.ennbb._si.d1_retained_ids[self.neighbors[i][j]]
                    self.pairs.append((entity_id, candidate_id, self.scores[i][j]))

        self.pairs = [(x[2], x[0], x[1]) for x in self.pairs]

    def _dfs_pair_emission(self) -> None:
        """Sorts NN neighborhoods in ascending average distance from their query entity,
           iterate over each neighborhoods' entities in ascending distance to query entity 
        """
        self.pairs = []

        average_neighborhood_distances = np.mean(self.scores, axis=1)
        sorted_neighborhoods = sorted_enumerate(average_neighborhood_distances)

        for sorted_neighborhood in sorted_neighborhoods:
            neighbor_scores = self.scores[sorted_neighborhood]
            neighbors = self.neighbors[sorted_neighborhood]
            entity_id = self.ennbb._si.d1_retained_ids[sorted_neighborhood] \
            if self.data.is_dirty_er \
            else self.ennbb._si.d2_retained_ids[sorted_neighborhood]
            
            for neighbor_index, neighbor in enumerate(neighbors):
                if(neighbor != -1):
                    neighbor_id = self.ennbb._si.d1_retained_ids[neighbor]
                    self.pairs.append((entity_id, neighbor_id, neighbor_scores[neighbor_index]))

        self.pairs = [(x[2], x[0], x[1]) for x in self.pairs]
        
    def _hb_pair_emission(self) -> None:
        """Sorts NN neighborhoods in ascending average distance from their query entity,
           emits the top entity for each neighborhood, then iterates over the sorte neighborhoods,
           and emits the pairs in descending weight order
        """
        self.pairs = []
        _first_emissions = []
        _remaining_emissions = []

        average_neighborhood_distances = np.mean(self.scores, axis=1)
        sorted_neighborhoods = sorted_enumerate(average_neighborhood_distances)

        for sorted_neighborhood in sorted_neighborhoods:
            neighbor_scores = self.scores[sorted_neighborhood]
            neighbors = self.neighbors[sorted_neighborhood]
            entity_id = self.ennbb._si.d1_retained_ids[sorted_neighborhood] \
            if self.data.is_dirty_er \
            else self.ennbb._si.d2_retained_ids[sorted_neighborhood]
            for neighbor_index, neighbor in enumerate(neighbors):
                if(neighbor != -1):
                    neighbor_id = self.ennbb._si.d1_retained_ids[neighbor]              
                    _current_emissions = _remaining_emissions if neighbor_index else _first_emissions
                    _current_emissions.append((entity_id, neighbor_id, neighbor_scores[neighbor_index]))

        self.pairs = [(x[2], x[0], x[1]) for x in _first_emissions] + [(x[2], x[0], x[1]) for x in _remaining_emissions]
        
    def _bfs_pair_emission(self) -> None:
        """Sorts NN neighborhoods in ascending average distance from their query entity,
           and iteratively emits the current top pair per neighborhood 
        """
        self.pairs = []
        average_neighborhood_distances = np.mean(self.scores, axis=1)
        sorted_neighborhoods = sorted_enumerate(average_neighborhood_distances)

        _emissions_per_pair = self.neighbors.shape[1]
        for current_emission_per_pair in range(_emissions_per_pair):
            for sorted_neighborhood in sorted_neighborhoods:
                neighbor = self.neighbors[sorted_neighborhood][current_emission_per_pair]
                if(neighbor != -1):
                    neighbor_id = self.ennbb._si.d1_retained_ids[neighbor]
                    entity_id = self.ennbb._si.d1_retained_ids[sorted_neighborhood] \
                    if self.data.is_dirty_er \
                    else self.ennbb._si.d2_retained_ids[sorted_neighborhood]
                    self.pairs.append((entity_id, neighbor_id, self.scores[sorted_neighborhood][current_emission_per_pair]))
                    
        self.pairs = [(x[2], x[0], x[1]) for x in self.pairs]

    def _produce_pairs(self):
        """Calls pairs emission based on the requested approach
        Raises:
            AttributeError: Given emission technique hasn't been defined
        """
        # currently first phase algorithms are in charge of gathering the subset of the original dataset
        # that will be used to initialize the scheduler, we simply retrieve all the pairs and their scores
        self._top_pair_emission()

    def _predict_raw_blocks(self, blocks: dict = None) -> List[Tuple[int, int]]:
        self.ennbb : EmbeddingsNNBlockBuilding = EmbeddingsNNBlockBuilding(self._language_model, self._similarity_search)
        
        self.final_blocks = self.ennbb.build_blocks(data=self.data,
                                                    vector_size=self._vector_size,
                                                    num_of_clusters=self._num_of_clusters,
                                                    top_k=self._number_of_nearest_neighbors,
                                                    return_vectors=False,
                                                    tqdm_disable=False,
                                                    save_embeddings=False,
                                                    load_embeddings_if_exist=False,
                                                    with_entity_matching=False,
                                                    input_cleaned_blocks=blocks,
                                                    similarity_distance=self.similarity_function)
        self.scores = self.ennbb.distances
        self.neighbors = self.ennbb.neighbors
        self.final_vectors = (self.ennbb.vectors_1, self.ennbb.vectors_2)
        self._produce_pairs()
        return self.pairs

    def _predict_prunned_blocks(self, blocks: dict = None) -> List[Tuple[int, int]]:
        return self._predict_raw_blocks(blocks)
    
class SimilarityBasedProgressiveMatching(ProgressiveMatching):
    """Applies similarity based candidate graph prunning, sorts retained comparisons and applies Progressive Matching
    """

    _method_name: str = "Similarity Based Progressive Matching"
    _method_info: str = "Applies similarity based candidate graph prunning, sorts retained comparisons and applies Progressive Matching"

    def __init__(
            self,
            weighting_scheme: str = 'ACF',
            window_size: int = 10,
            similarity_function: str = 'dice',
            tokenizer: str = 'white_space_tokenizer',
            vectorizer : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.0,
            tokenizer_return_unique_values = True, # unique values or not
            attributes: any = None
        ) -> None:
        super().__init__(similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)
        self._weighting_scheme : str = weighting_scheme
        self._window_size : int = window_size
        
class GlobalPSNM(SimilarityBasedProgressiveMatching):
    """Applies Global Progressive Sorted Neighborhood Matching
    """

    _method_name: str = "Global Progressive Sorted Neighborhood Matching"
    _method_info: str = "For each entity sorted accordingly to its block's tokens, " + \
                        "evaluates its neighborhood pairs defined within shifting windows of incremental size" + \
                        " and retains the globally best candidate pairs"

    def __init__(
            self,
            weighting_scheme: str = 'ACF',
            window_size: int = 10,
            similarity_function: str = 'dice',
            tokenizer: str = 'white_space_tokenizer',
            vectorizer : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.0,
            tokenizer_return_unique_values = True, # unique values or not
            attributes: any = None
        ) -> None:

        super().__init__(weighting_scheme=weighting_scheme,
                        window_size=window_size,
                        similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)

    def _predict_raw_blocks(self, blocks: dict) -> List[Tuple[float, int, int]]:
        gpsn : GlobalProgressiveSortedNeighborhood = GlobalProgressiveSortedNeighborhood(self._weighting_scheme, self._budget)
        self.pairs : List[Tuple[float, int, int]] = gpsn.process(blocks=blocks, data=self.data, window_size=self._window_size, tqdm_disable=True, emit_all_tps_stop=self._emit_all_tps_stop)
        return self.pairs

    def _predict_prunned_blocks(self, blocks: dict) -> List[Tuple[float, int, int]]:
        raise NotImplementedError("Sorter Neighborhood Algorithms don't support prunned blocks")
    
class LocalPSNM(SimilarityBasedProgressiveMatching):
    """Applies Local Progressive Sorted Neighborhood Matching
    """

    _method_name: str = "Global Progressive Sorted Neighborhood Matching"
    _method_info: str = "For each entity sorted accordingly to its block's tokens, " + \
                        "evaluates its neighborhood pairs defined within shifting windows of incremental size" + \
                        " and retains the globally best candidate pairs"

    def __init__(
            self,
            weighting_scheme: str = 'ACF',
            window_size: int = 10,
            similarity_function: str = 'dice',
            tokenizer: str = 'white_space_tokenizer',
            vectorizer : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.0,
            tokenizer_return_unique_values = True, # unique values or not
            attributes: any = None
        ) -> None:

        super().__init__(weighting_scheme=weighting_scheme,
                        window_size=window_size,
                        similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)

    def _predict_raw_blocks(self, blocks: dict) -> List[Tuple[float, int, int]]:
        lpsn : LocalProgressiveSortedNeighborhood = LocalProgressiveSortedNeighborhood(self._weighting_scheme, self._budget)
        self.pairs : List[Tuple[float, int, int]] = lpsn.process(blocks=blocks, data=self.data, window_size=self._window_size, tqdm_disable=True, emit_all_tps_stop=self._emit_all_tps_stop)
        return self.pairs

    def _predict_prunned_blocks(self, blocks: dict) -> List[Tuple[float, int, int]]:
        raise NotImplementedError("Sorter Neighborhood Algorithms don't support prunned blocks " + \
                                "(pre comparison-cleaning entities per block distribution required")
class RandomPM(ProgressiveMatching):
    """Picks a number of random comparisons equal to the available budget
    """

    _method_name: str = "Random Progressive Matching"
    _method_info: str = "Picks a number of random comparisons equal to the available budget"

    def __init__(
            self,
            similarity_function: str = 'dice',
            tokenizer: str = 'white_space_tokenizer',
            similarity_threshold: float = 0.0,
            qgram: int = 2, # for jaccard
            tokenizer_return_unique_values = True, # unique values or not
            attributes: any = None,
            delim_set: list = None, # DelimiterTokenizer
            padding: bool = True, # QgramTokenizer
            prefix_pad: str = '#', # QgramTokenizer (if padding=True)
            suffix_pad: str = '$' # QgramTokenizer (if padding=True)
        ) -> None:

        super().__init__(similarity_function, tokenizer, similarity_threshold, qgram, tokenizer_return_unique_values, attributes, delim_set, padding, prefix_pad, suffix_pad)

    def _predict_raw_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        cp : ComparisonPropagation = ComparisonPropagation()
        cleaned_blocks = cp.process(blocks=blocks, data=self.data, tqdm_disable=True)
        self._predict_prunned_blocks(cleaned_blocks)

    def _predict_prunned_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        _all_pairs = [(id1, id2) for id1 in blocks for id2 in blocks[id1]]
        _total_pairs = len(_all_pairs)
        random_pairs = sample(_all_pairs, self._budget) if self._budget <= _total_pairs and not self._emit_all_tps_stop else _all_pairs
        self.pairs.add_edges_from(random_pairs)
        
class PESM(HashBasedProgressiveMatching):
    """Applies Progressive Entity Scheduling Matching
    """

    _method_name: str = "Progressive Entity Scheduling Matching"
    _method_info: str = "Applies Progressive Entity Scheduling - Sorts entities in descending order of their average weight, " + \
                        "emits the top pair per entity. Finally, traverses the sorted " + \
                        "entities and emits their comparisons in descending weight order " + \
                        "within specified budget."
                        
    def __init__(
            self,
            weighting_scheme: str = 'CBS',
            similarity_function: str = 'dice',
            tokenizer: str = 'white_space_tokenizer',
            vectorizer : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.0,
            tokenizer_return_unique_values = True, # unique values or not
            attributes: any = None,
        ) -> None:

        super().__init__(weighting_scheme=weighting_scheme,
                        similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=vectorizer,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)

    def _predict_raw_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        
        pes : ProgressiveEntityScheduling = ProgressiveEntityScheduling(self._weighting_scheme, self._budget)
        self.pairs = pes.process(blocks=blocks, data=self.data, tqdm_disable=True, cc=None, method=self._algorithm, emit_all_tps_stop=self._emit_all_tps_stop)
        return self.pairs
    
    def _predict_prunned_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        return self._predict_raw_blocks(blocks)
        # raise NotImplementedError("Sorter Neighborhood Algorithms doesn't support prunned blocks (lack of precalculated weights)")
    
# class WhooshPM(BlockIndependentPM):
#     """Applies progressive index based matching using whoosh library 
#     """

#     _method_name: str = "Whoosh Progressive Matching"
#     _method_info: str = "Applies Whoosh Progressive Matching - Indexes the entities of the second dataset, " + \
#                         "stores their specified attributes, " + \
#                         "defines a query for each entity of the first dataset, " + \
#                         "and retrieves its pair candidates from the indexer within specified budget"

#     def __init__(
#             self,
#             similarity_function: str = 'WH-TF-IDF',
#             number_of_nearest_neighbors: int = 10,
#             tokenizer: str = 'white_space_tokenizer',
#             similarity_threshold: float = 0.0,
#             qgram: int = 2, # for jaccard
#             tokenizer_return_unique_values = True, # unique values or not
#             attributes: any = None,
#             delim_set: list = None, # DelimiterTokenizer
#             padding: bool = True, # QgramTokenizer
#             prefix_pad: str = '#', # QgramTokenizer (if padding=True)
#             suffix_pad: str = '$' # QgramTokenizer (if padding=True)
#         ) -> None:
#         # budget set to float('inf') implies unlimited budget
#         super().__init__(similarity_function, tokenizer, similarity_threshold, qgram, tokenizer_return_unique_values, attributes, delim_set, padding, prefix_pad, suffix_pad)
#         self._number_of_nearest_neighbors : int = number_of_nearest_neighbors
        
#     def _set_whoosh_datasets(self) -> None:
#         """Saves the rows of both datasets corresponding to the indices of the entities that have been retained after comparison cleaning
#         """
        
#         self._whoosh_d1 = self.data.dataset_1[self.attributes + [self.data.id_column_name_1]] if self.attributes else self.data.dataset_1
#         self._whoosh_d1 = self._whoosh_d1[self._whoosh_d1[self.data.id_column_name_1].isin(self._whoosh_d1_retained_index)]
#         if(not self.data.is_dirty_er):  
#             self._whoosh_d2 = self.data.dataset_2[self.attributes + [self.data.id_column_name_2]] if self.attributes else self.data.dataset_2
#             self._whoosh_d2 = self._whoosh_d2[self._whoosh_d2[self.data.id_column_name_2].isin(self._whoosh_d2_retained_index)]
        

#     def _set_retained_entries(self) -> None:
#         """Saves the indices of entities of both datasets that have been retained after comparison cleaning
#         """
#         self._whoosh_d1_retained_index = pd.Index([self.data._gt_to_ids_reversed_1[id] 
#         for id in self._si.d1_retained_ids])
        
#         if(not self.data.is_dirty_er):
#             self._whoosh_d2_retained_index = pd.Index([self.data._gt_to_ids_reversed_2[id] 
#         for id in self._si.d2_retained_ids])
    
    
#     def _initialize_index_path(self):
#         """Creates index directory if non-existent, constructs the absolute path to the current whoosh index
#         """
#         global INDEXER_DIR
#         INDEXER_DIR = os.path.abspath(INDEXER_DIR)  
#         _d1_name = self.data.dataset_name_1 if self.data.dataset_name_1 is not None else 'd3'    
#         self._index_path = os.path.join(INDEXER_DIR, _d1_name if self.data.is_dirty_er else (_d1_name + (self.data.dataset_name_2 if self.data.dataset_name_2 is not None else 'd4')))
#         if not os.path.exists(self._index_path):
#             print('Created index directory at: ' + self._index_path)
#             os.makedirs(self._index_path, exist_ok=True)
        
    
#     def _create_index(self):
#         """Defines the schema [ID, CONTENT], creates the index in the defined path 
#            and populates it with all the entities of the target dataset (first - Dirty ER, second - Clean ER)
#         """
#         self._schema = Schema(ID=ID(stored=True), content=TEXT(stored=True))
#         self._index = create_in(self._index_path, self._schema)
#         writer = self._index.writer()
        
#         _target_dataset = self._whoosh_d1 if self.data.is_dirty_er else self._whoosh_d2
#         _id_column_name = self.data.id_column_name_1 if self.data.is_dirty_er else self.data.id_column_name_2
        
#         for _, entity in _target_dataset.iterrows():
#             entity_values = [str(entity[column]) for column in _target_dataset.columns if column != _id_column_name]
#             writer.add_document(ID=entity[_id_column_name], content=' '.join(entity_values))
#         writer.commit()
    
#     def _populate_whoosh_dataset(self) -> None:
#         """For each retained entity in the first dataset, construct a query with its text content,
#            parses it to the indexers, retrieves best candidates and stores them in entity's neighborhood.
#            Populates a list with all the retrieved pairs.
#         """
#         # None value for budget implies unlimited budget in whoosh 
#         _query_budget = self._number_of_nearest_neighbors
        
#         if(self.similarity_function not in whoosh_similarity_function):
#             print(f'{self.similarity_function} Similarity Function is Undefined')
#             self.similarity_function = 'Frequency'
#         print(f'Applying {self.similarity_function} Similarity Function')
#         _scorer = whoosh_similarity_function[self.similarity_function]
        
#         with self._index.searcher(weighting=_scorer) as searcher:
#             self._parser = qparser.QueryParser('content', schema=self._index.schema, group=qparser.OrGroup)
#             for _, entity in self._whoosh_d1.iterrows():
#                 entity_values = [str(entity[column]) for column in self._whoosh_d1.columns if column != self.data.id_column_name_1]
#                 entity_string = ' '.join(entity_values)
#                 entity_id = entity[self.data.id_column_name_1]
#                 entity_query = self._parser.parse(entity_string)
#                 query_results = searcher.search(entity_query, limit = _query_budget)
                
#                 for neighbor in query_results:
#                     _score = neighbor.score
#                     _neighbor_id = neighbor['ID']
#                     self.pairs.append((_score, self.data._ids_mapping_1[entity], self.data._ids_mapping_2[_neighbor_id]))
                       
#     def _predict_raw_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
#         self._start_time = time()
#         self._si = SubsetIndexer(blocks=blocks, data=self.data, subset=False)
#         self._set_retained_entries()
#         self._set_whoosh_datasets()
#         self._initialize_index_path()
#         self._create_index()
#         self.pairs : List[Tuple[float, int, int]] = []
#         self._budget = float('inf') if self._emit_all_tps_stop else self._budget
#         self._populate_whoosh_dataset()
#         self.execution_time = time() - self._start_time
#         return self.pairs
        
#     def _predict_prunned_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
#         self._predict_raw_blocks(blocks)  

class TopKJoinPM(ProgressiveMatching):
    """Applies index based matching for ES, emits candidate pairs using defined budget/emission technique
    """

    _method_name: str = "Top-K Join Progressive Matching"
    _method_info: str = "Applies index based matching for ES, emits candidate pairs using defined budget/emission technique"
    def __init__(
            self,
            similarity_function: str = 'dice',
            number_of_nearest_neighbors : int = 10,
            tokenizer: str = None,
            weighting_scheme : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.0,
            tokenizer_return_unique_values = True, # unique values or not
            attributes: any = None,
        ) -> None:

        super().__init__(similarity_function=similarity_function,
                        tokenizer=tokenizer,
                        vectorizer=weighting_scheme,
                        qgram=qgram,
                        similarity_threshold=similarity_threshold,
                        tokenizer_return_unique_values=tokenizer_return_unique_values,
                        attributes=attributes)
        
        self.similarity_function : str = similarity_function
        self.number_of_nearest_neighbors : int = number_of_nearest_neighbors
        self.weighting_scheme : str = weighting_scheme
        self.qgram : int = qgram 
        
    def _predict_raw_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        ptkj : TopKJoin = TopKJoin(K=self.number_of_nearest_neighbors,
                                   metric=self.similarity_function,
                                   tokenization=self.tokenizer,
                                   qgrams=self.qgram)
        if(self.weighting_scheme is not None):
            ptkj.vectorizer = self.initialize_vectorizer()
        self.pairs = ptkj.fit(data=self.data,
                              reverse_order=(self._indexing=='reverse'),
                              attributes_1=self.data.attributes_1,
                              attributes_2=self.data.attributes_2)
        self.pairs = [(edge[2]['weight'], edge[0], edge[1]) for edge in self.pairs.edges(data=True)]
        return self.pairs

    
    def _predict_prunned_blocks(self, blocks: dict) -> List[Tuple[int, int]]:
        raise NotImplementedError("Progressive TopKJoin PM for prunned blocks - Not implemented yet!")
        
    
        
    def initialize_vectorizer(self) -> FrequencyEvaluator:
        self.vectorizer : FrequencyEvaluator = FrequencyEvaluator(vectorizer=self.weighting_scheme,
                                                                  tokenizer=self.tokenizer,
                                                                  qgram=self.qgram)
        
        d1 = self.data.dataset_1[self.data.attributes_1] if self.data.attributes_1 is not None else self.data.dataset_1
        self._entities_d1 = d1 \
                    .apply(" ".join, axis=1) \
                    .apply(lambda x: x.lower()) \
                    .values.tolist()
        d2 = self.data.dataset_2[self.data.attributes_2] if self.data.attributes_2 is not None else self.data.dataset_2
        self._entities_d2 = d2 \
                    .apply(" ".join, axis=1) \
                    .apply(lambda x: x.lower()) \
                    .values.tolist() if not self.data.is_dirty_er else None
        
        self.vectorizer.fit(metric=self.similarity_function,
                            d1_entities=self._entities_d1,
                            d2_entities=self._entities_d2)
        
        return self.vectorizer

class_references = {
    'GlobalTopPM' : GlobalTopPM,
    'LocalTopPM' : LocalTopPM,
    'GlobalPSNM' : GlobalPSNM,
    'LocalPSNM' : LocalPSNM,
    'PESM' : PESM,
    'EmbeddingsNNBPM' : EmbeddingsNNBPM,
    'TopKJoinPM' : TopKJoinPM
} 

